"""Models and functions for CLI validation."""
import argparse

from ywh2bt.cli.error import CliError
from ywh2bt.core.configuration.error import BaseAttributeError
from ywh2bt.core.core import load_configuration


def validate(
    args: argparse.Namespace,
) -> None:
    """
    Validate a configuration file.

    Args:
        args: command line arguments

    Raises:
        CliError: if an error occurred
    """
    configuration = load_configuration(
        config_format=args.config_format,
        config_file=args.config_file,
    )
    try:
        configuration.validate()
    except BaseAttributeError as e:
        raise CliError("Invalid configuration") from e



"""Models and functions used in schema documentation dialogs."""
from typing import (
    Any,
    Optional,
)

import markdown
from PySide6.QtGui import (
    QAction,
    QKeySequence,
)
from PySide6.QtWebEngineCore import QWebEnginePage
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtWidgets import (
    QMainWindow,
    QWidget,
)

from ywh2bt.core.schema.markdown import root_configuration_as_markdown
from ywh2bt.gui.widgets.typing import as_signal_instance


class SchemaDocumentationDialog(QMainWindow):
    """A window that display the schema documentation."""

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
        )
        self.setWindowTitle("ywh2bt - Schema documentation")
        self._init_ui()

    def _init_ui(
        self,
    ) -> None:
        self._create_menus()
        self.setCentralWidget(self._create_view())
        self.statusBar().show()

    def _create_menus(
        self,
    ) -> None:
        bar = self.menuBar()

        file_menu = bar.addMenu("&File")
        file_menu.setObjectName("menu_file")
        file_menu.addAction(self._create_close_action())

    def _create_close_action(
        self,
    ) -> QAction:
        action = QAction("&Close", self)
        action.setObjectName("action_close")
        action.setShortcuts(QKeySequence.StandardKey.Close)
        action.setStatusTip("Close")
        as_signal_instance(action.triggered).connect(
            self._on_close_triggered,
        )
        return action

    def _on_close_triggered(
        self,
    ) -> None:
        self.close()

    def _create_view(
        self,
    ) -> QWebEngineView:
        page = QWebEnginePage(self)
        page.setHtml(
            markdown.markdown(
                root_configuration_as_markdown(),
            ),
        )
        view = QWebEngineView(self)
        view.setPage(page)
        return view



"""Models and functions used for exportable list GUI."""
from __future__ import annotations

from functools import partial
from typing import (
    Any,
    List,
    Optional,
    Set,
    Union,
)

from PySide6.QtCore import (
    QAbstractTableModel,
    QItemSelection,
    QModelIndex,
    QPersistentModelIndex,
    Qt,
    Signal,
)
from PySide6.QtWidgets import (
    QAbstractItemView,
    QHBoxLayout,
    QHeaderView,
    QLayout,
    QPushButton,
    QTableView,
    QVBoxLayout,
    QWidget,
)

from ywh2bt.core.configuration.attribute import ExportableList
from ywh2bt.gui.widgets import constants
from ywh2bt.gui.widgets.typing import as_signal_instance


class ExportableListWidget(QWidget):
    """Exportable list GUI."""

    dataChanged: Signal = Signal(ExportableList)  # noqa: N815

    _table: QTableView
    _add_button: QPushButton
    _del_button: QPushButton

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._init_ui()

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._table = self._create_table()

        self._add_button = self._create_add_button()
        self._del_button = self._create_del_button()

        buttons_layout = self._create_buttons_layout(
            self._add_button,
            self._del_button,
        )

        layout.addWidget(self._table)
        layout.addLayout(buttons_layout)
        layout.setContentsMargins(0, 0, 0, 0)

    def _create_table(
        self,
    ) -> QTableView:
        widget = QTableView(self)
        widget.horizontalHeader().hide()
        widget.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        widget.setEditTriggers(
            QAbstractItemView.EditTrigger.AllEditTriggers,
        )
        return widget

    def _create_add_button(
        self,
    ) -> QPushButton:
        widget = QPushButton(
            "+",
            self,
        )
        widget.setFixedSize(constants.SMALL_BUTTON_SIZE)
        as_signal_instance(widget.clicked).connect(
            self._on_add_button_clicked,
        )
        return widget

    def _create_del_button(
        self,
    ) -> QPushButton:
        widget = QPushButton(
            "-",
            self,
        )
        widget.setFixedSize(constants.SMALL_BUTTON_SIZE)
        widget.setEnabled(False)
        as_signal_instance(widget.clicked).connect(
            self._on_del_button_clicked,
        )
        return widget

    def _create_buttons_layout(
        self,
        *buttons: QPushButton,
    ) -> QLayout:
        layout = QVBoxLayout()
        for button in buttons:
            layout.addWidget(button, 1)
        layout.addStretch(1)
        return layout

    def set_model(
        self,
        model: ExportableListModel,
    ) -> None:
        """
        Set the model.

        Args:
            model: an exportable list model
        """
        signals = {
            model.dataChanged,
            model.rowsInserted,
            model.rowsRemoved,
        }
        for signal in signals:
            as_signal_instance(signal).connect(
                partial(
                    self._on_model_data_changed,
                    model,
                ),
            )
        self._table.setModel(model)
        selection_model = self._table.selectionModel()
        as_signal_instance(selection_model.selectionChanged).connect(
            self._on_selection_changed,
        )

    def _on_model_data_changed(
        self,
        model: ExportableListModel,
        top_left: QModelIndex,
        bottom_right: QModelIndex,
        roles: List[int],
    ) -> None:
        as_signal_instance(self.dataChanged).emit(model.get_value())

    def _on_selection_changed(
        self,
        selected: QItemSelection,
        unselected: QItemSelection,
    ) -> None:
        has_selection = selected.count() != 0
        self._del_button.setEnabled(has_selection)

    def _on_add_button_clicked(
        self,
        checked: bool,
    ) -> None:
        model = self._table.model()
        model.insertRow(model.rowCount(model.index(0, 0)))
        row_count = model.rowCount(model.index(0, 0))
        self._table.selectRow(max(0, row_count - 1))

    def _on_del_button_clicked(
        self,
        checked: bool,
    ) -> None:
        selected_rows = sorted(
            self._get_select_row_indices(),
            reverse=True,
        )
        if not selected_rows:
            return
        model = self._table.model()
        for selected_row in selected_rows:
            model.removeRow(selected_row)

        row_count = model.rowCount(model.index(0, 0))
        reselect_row = min(row_count - 1, selected_rows[-1])
        self._table.selectRow(max(0, reselect_row))

    def _get_select_row_indices(
        self,
    ) -> Set[int]:
        selection_model = self._table.selectionModel()
        return set({index.row() for index in selection_model.selectedIndexes()})


class ExportableListModel(QAbstractTableModel):
    """Exportable list model."""

    _exportable_list: ExportableList[Any, Any]

    def __init__(
        self,
        exportable_list: ExportableList[Any, Any],
    ) -> None:
        """
        Initialize self.

        Args:
            exportable_list: an exportable list
        """
        super().__init__()
        self._exportable_list = exportable_list

    def get_value(
        self,
    ) -> Optional[ExportableList[Any, Any]]:
        """
        Get the underlying exportable list.

        Returns:
            The exportable list
        """
        return self._exportable_list

    def data(  # type: ignore
        self,
        index: QModelIndex,
        role: int,
    ) -> Any:
        """
        Get data at the given index for the given role.

        Args:
            index: a data index
            role: a data role

        Returns:
            The data
        """
        return_conditions = (
            self._exportable_list is None,
            not index.isValid(),
            index.column() != 0,
        )
        if any(return_conditions):
            return None
        if role in {Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole}:
            return self._exportable_list[index.row()]
        return None

    def setData(  # type: ignore # noqa: N802
        self,
        index: QModelIndex,
        value: Any,
        role: int,
    ) -> bool:
        """
        Set data at the given index for the given role.

        Args:
            index: a data index
            value: a value
            role: a role

        Returns:
            True if the data has been set, otherwise False
        """
        if self._exportable_list is None or not index.isValid() or role != Qt.ItemDataRole.EditRole:
            return False
        if self._exportable_list[index.row()] == value:
            return False
        self._exportable_list[index.row()] = value
        self.dataChanged.emit(index, index)
        return True

    def rowCount(  # type: ignore # noqa: N802
        self,
        index: QModelIndex,
    ) -> int:
        """
        Get the number of rows/entries in the underlying exportable list.

        Args:
            index: a data index for the parent data, if any

        Returns:
            The number of rows
        """
        return len(self._exportable_list) if self._exportable_list is not None else 0

    def columnCount(  # type: ignore # noqa: N802
        self,
        index: Union[QModelIndex, QPersistentModelIndex],
    ) -> int:
        """
        Get the number of columns in the underlying exportable list, which is always 1.

        Args:
            index: a data index for the parent data, if any

        Returns:
            The number of columns
        """
        return 1

    def flags(
        self,
        index: Union[QModelIndex, QPersistentModelIndex],
    ) -> Qt.ItemFlag:
        """
        Get the flags for the given index.

        Args:
            index: a data index

        Returns:
            The flags
        """
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled
        return super().flags(index) | Qt.ItemFlag.ItemIsEditable

    def insertRows(  # type: ignore # noqa: N802
        self,
        row: int,
        count: int,
        parent: QModelIndex,
    ) -> bool:
        """
        Insert a number of rows before a given row.

        If row is 0, the rows are prepended to any existing rows in the parent.

        If row is rowCount(), the rows are appended to any existing rows in the parent.

        If parent has no children, a single column with count rows is inserted.

        Args:
            row: a row number
            count: a number of rows to be inserted
            parent: a parent index

        Returns:
            True if the rows where successfully inserted, otherwise False
        """
        self.beginInsertRows(
            parent,
            row,
            row + count - 1,
        )
        current_count = len(self._exportable_list)
        for i in range(count):
            value = self._get_next_new_available_entry(
                base_index=i,
            )
            if row == 0:
                self._exportable_list.insert(i, value)
            elif row == current_count:
                self._exportable_list.append(value)
        self.endInsertRows()
        return True

    def _get_next_new_available_entry(
        self,
        base_index: int,
    ) -> str:
        n = 1
        while True:
            value = f"New entry {base_index + n}"
            if value not in self._exportable_list:
                return value
            n += 1

    def removeRows(  # type: ignore # noqa: N802
        self,
        row: int,
        count: int,
        parent: QModelIndex,
    ) -> bool:
        """
        Remove count rows starting with the given row under parent.

        Args:
            row: a row number
            count: a number of rows
            parent: a parent index

        Returns:
            True if the rows were successfully removed; otherwise False.
        """
        self.beginRemoveRows(
            parent,
            row,
            row + count - 1,
        )
        for _ in range(count):
            self._exportable_list.pop(row)
        self.endRemoveRows()
        return True



"""Models and functions used for manipulating html data."""
import re
from copy import deepcopy
from html import escape as html_escape
from html import unescape as html_unescape
from typing import (
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    cast,
)
from urllib.parse import (
    parse_qs,
    unquote,
    urlsplit,
)

import html2text
from bs4 import BeautifulSoup  # type: ignore

from ywh2bt.core.api.models.report import Attachment


def ywh_html_to_markdown(
    html: str,
) -> str:
    """
    Convert html emanating from YWH to markdown.

    Args:
        html: an html

    Returns:
        a markdown
    """
    markdown = html_to_markdown(html)
    offset = 0
    languages = _extract_language_codes(
        html=html,
    )
    for language in languages:
        offset = markdown[offset:].index("[code]") + 6 + offset
        markdown = markdown[:offset] + language + markdown[offset:]
    return markdown.replace("[code]", "```",).replace(
        "[/code]",
        "```",
    )


def _extract_language_codes(
    html: str,
) -> Iterator[str]:
    soup = BeautifulSoup(html, "lxml")
    for pre in soup.findAll("pre"):
        for code in pre:
            class_attribute = code.attrs.get("class", [""])[0]
            yield class_attribute.replace("language-", "")


def html_to_markdown(
    html: str,
) -> str:
    """
    Convert html to markdown.

    Args:
        html: an html

    Returns:
        a markdown
    """
    html_parser = html2text.HTML2Text()
    html_parser.body_width = 0
    html_parser.mark_code = True
    return html_parser.handle(html)


def cleanup_ywh_redirects_from_html(
    ywh_domain: str,
    html: str,
) -> str:
    """
    Replace YesWeHack redirects with real URLs.

    Args:
        ywh_domain: a base domain of the YWH redirects
        html: an html

    Returns:
        the cleaned html
    """
    redirect_base_re = re.escape(f"{ywh_domain}/redirect?url=")

    pattern = re.compile(f'"(https?://{redirect_base_re}[^ "]*)"')
    redirect_urls = pattern.findall(html)
    for redirect_url in redirect_urls:
        real_url = _extract_real_url_from_redirect(
            redirect_url=html_unescape(redirect_url),
        )
        html = html.replace(
            redirect_url,
            html_escape(real_url or ""),
        )
    return html


def cleanup_ywh_redirects_from_text(
    ywh_domain: str,
    text: str,
) -> str:
    """
    Replace YesWeHack redirects with real URLs.

    Args:
        ywh_domain: a base domain of the YWH redirects
        text: an plain text

    Returns:
        the cleaned text
    """
    redirect_base_re = re.escape(f"{ywh_domain}/redirect?url=")

    pattern = re.compile(rf"(https?://{redirect_base_re}\S*)\b")
    redirect_urls = pattern.findall(text)
    for redirect_url in redirect_urls:
        real_url = _extract_real_url_from_redirect(
            redirect_url=redirect_url,
        )
        text = text.replace(
            redirect_url,
            real_url or "",
        )
    return text


def _extract_real_url_from_redirect(
    redirect_url: str,
) -> Optional[str]:
    parse_result = urlsplit(unquote(redirect_url))
    params = parse_qs(parse_result.query)
    if "url" in params:
        return cast(List[str], params.get("url"))[0]
    return None


def cleanup_attachments_and_urls_from_html(
    html: str,
    attachments: List[Attachment],
) -> Tuple[str, List[Attachment]]:
    """
    Remove query parameters from attachments urls and from references in the HTML.

    Args:
        html: an html
        attachments: a list of attachments

    Returns:
        the cleaned html
    """
    if not attachments:
        return html, attachments
    clean_attachments = []
    clean_attachments_urls = []
    attachments_domains: Set[str] = set()
    for attachment in attachments:
        parsed_url = urlsplit(attachment.url)
        attachments_domains.add(parsed_url.netloc)
        clean_attachment = deepcopy(attachment)
        clean_attachment.url = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"
        clean_attachments.append(clean_attachment)
        clean_attachments_urls.append(clean_attachment.url)
    domains_base_re = "|".join([re.escape(domain) for domain in attachments_domains])
    pattern = re.compile(f'"(https?://(?:{domains_base_re})[^ "]*)"')
    html_urls = pattern.findall(html)
    for html_url in html_urls:
        parsed_html_url = urlsplit(html_url)
        clean_html_url = f"{parsed_html_url.scheme}://{parsed_html_url.netloc}{parsed_html_url.path}"
        if clean_html_url in clean_attachments_urls:
            html = html.replace(
                html_url,
                clean_html_url,
            )
    return html, clean_attachments



"""Models and functions used in data hashing."""
from PySide6.QtCore import (
    QByteArray,
    QCryptographicHash,
    QFile,
)


def file_checksum(
    file_path: str,
    algorithm: QCryptographicHash.Algorithm = QCryptographicHash.Algorithm.Sha256,
) -> QByteArray:
    """
    Compute a checksum of the contents of a file.

    Args:
        file_path: a path to a file
        algorithm: an algorithm used to compute the hash

    Returns:
        The checksum
    """
    f = QFile(file_path)
    if f.open(QFile.OpenModeFlag.ReadOnly):
        h = QCryptographicHash(algorithm)
        if h.addData(f):
            return h.result()
    return QByteArray()



"""Models and functions used for YesWeHack api clients factories."""
from abc import (
    ABC,
    abstractmethod,
)
from typing import Dict

from ywh2bt.core.api.yeswehack import (
    YesWeHackApiClient,
    YesWeHackApiClientError,
)
from ywh2bt.core.configuration.yeswehack import YesWeHackConfiguration
from ywh2bt.core.exceptions import CoreException


class YesWeHackApiClientsAbstractFactory(ABC):
    """Abstract factory for YesWeHackApiClients."""

    @abstractmethod
    def get_yeswehack_api_client(
        self,
        configuration: YesWeHackConfiguration,
    ) -> YesWeHackApiClient:
        """
        Get the api client for the given configuration.

        Args:
            configuration: a configuration
        """


class YesWeHackApiClientsFactory(YesWeHackApiClientsAbstractFactory):
    """Concrete factory for YesWeHackApiClients."""

    _yeswehack_api_clients: Dict[YesWeHackConfiguration, YesWeHackApiClient]

    def __init__(
        self,
    ) -> None:
        """Initialize self."""
        self._yeswehack_api_clients = {}

    def get_yeswehack_api_client(
        self,
        configuration: YesWeHackConfiguration,
    ) -> YesWeHackApiClient:
        """
        Get the api client for the given configuration.

        Args:
            configuration: a configuration

        Raises:
            CoreException: if the client could not be created

        Returns:
            The client
        """
        if configuration not in self._yeswehack_api_clients:
            try:
                client = YesWeHackApiClient(
                    configuration=configuration,
                )
            except YesWeHackApiClientError as e:
                raise CoreException("Unable to create YesWeHack API client") from e
            self._yeswehack_api_clients[configuration] = client
        return self._yeswehack_api_clients[configuration]



"""Models and functions used for schema."""
from typing_extensions import Protocol


class SchemaDumpProtocol(Protocol):
    """Protocol for schema dump."""

    def __call__(
        self,
    ) -> str:
        """Dump the schema."""
        ...



"""Models and functions used for representing states."""

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
)


class State(ABC):
    """A state."""

    @abstractmethod
    def as_dict(self) -> Dict[str, Any]:
        """Get the fields as a new dict."""



"""Models and functions used for attributes container GUI."""
from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import (
    Any,
    Dict,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from PySide6.QtCore import (
    Qt,
    Signal,
)
from PySide6.QtWidgets import (
    QFormLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)
from typing_extensions import Protocol

from ywh2bt.core.configuration.attribute import (
    Attribute,
    AttributesContainer,
    AttributesContainerDict,
    AttributesContainerList,
    ExportableDict,
    ExportableList,
)
from ywh2bt.gui.widgets.attribute.attributes_container_dict_widget import AttributesContainerDictWidget
from ywh2bt.gui.widgets.attribute.attributes_container_list_widget import AttributesContainerListWidget
from ywh2bt.gui.widgets.attribute.exportable_dict_widget import (
    ExportableDictModel,
    ExportableDictWidget,
)
from ywh2bt.gui.widgets.attribute.exportable_list_widget import (
    ExportableListModel,
    ExportableListWidget,
)
from ywh2bt.gui.widgets.hinted_check_box_widget import HintedCheckBoxWidget
from ywh2bt.gui.widgets.secret_line_edit_widget import SecretLineEditWidget
from ywh2bt.gui.widgets.typing import as_signal_instance


T = TypeVar("T")


class _WidgetCreationProtocol(Protocol):
    def __call__(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> QWidget:
        ...


class _WidgetUpdateProtocol(Protocol):
    def __call__(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        ...


@dataclass
class _WidgetProtocols:
    create: _WidgetCreationProtocol
    update: _WidgetUpdateProtocol


class AttributesContainerWidget(QWidget):
    """A widget for editing AttributesContainer instances."""

    container_changed: Signal = Signal(AttributesContainer)
    _container_class: Type[AttributesContainer]
    _container: Optional[AttributesContainer]

    _widget_types_protocols: Dict[
        Type[Any],
        _WidgetProtocols,
    ]

    def __init__(
        self,
        container_class: Type[AttributesContainer],
        container: Optional[AttributesContainer] = None,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            container_class: a class extending AttributesContainer
            container: a container of type container_class
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._container_class = container_class
        self._container = None
        self._widget_types_protocols = {
            AttributesContainer: _WidgetProtocols(
                create=self._create_attributes_container_widget,
                update=self._update_attributes_container_widget,
            ),
            AttributesContainerDict: _WidgetProtocols(
                create=self._create_attributes_container_dict_widget,
                update=self._update_attributes_container_dict_widget,
            ),
            AttributesContainerList: _WidgetProtocols(
                create=self._create_attributes_container_list_widget,
                update=self._update_attributes_container_list_widget,
            ),
            ExportableDict: _WidgetProtocols(
                create=self._create_exportable_dict_widget,
                update=self._update_exportable_dict_widget,
            ),
            ExportableList: _WidgetProtocols(
                create=self._create_exportable_list_widget,
                update=self._update_exportable_list_widget,
            ),
            str: _WidgetProtocols(
                create=self._create_line_edit_widget,
                update=self._update_line_edit_widget,
            ),
            bool: _WidgetProtocols(
                create=self._create_check_box_widget,
                update=self._update_check_box_widget,
            ),
        }
        self._init_ui()
        self.set_container(
            container=container,
        )

    def set_container(
        self,
        container: Optional[AttributesContainer],
    ) -> None:
        """
        Set the container to be edited.

        Args:
            container: a container
        """
        self._container = container
        for name, attribute in self._container_class.get_attributes().items():
            value = getattr(container, name) if container else None
            self._update_attribute_widget(
                name=name,
                attribute=attribute,
                value=value,
            )

    def _init_ui(self) -> None:
        layout = QVBoxLayout(self)

        form_layout = self._create_form_layout()
        layout.addLayout(form_layout, 1)
        layout.addStretch(1)

    def _create_form_layout(
        self,
    ) -> QFormLayout:
        layout = QFormLayout()
        layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        for name, attribute in self._container_class.get_attributes().items():
            label = self._create_label_widget(
                name=name,
                attribute=attribute,
            )
            field = self._create_field_widget(
                name=name,
                attribute=attribute,
            )
            tool_tip = self._get_tool_tip(
                attribute=attribute,
            )
            if tool_tip:
                label.setToolTip(tool_tip)
                label.setStatusTip(tool_tip)

            layout.addRow(label, field)
        return layout

    def _create_label_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> QLabel:
        widget = QLabel(self)
        widget.setTextFormat(Qt.TextFormat.RichText)
        text = attribute.short_description or name
        text = f"{text}:"
        if attribute.deprecated:
            text = f"<i><s>{text}</s></i>"
        if attribute.required:
            text = f"<b>{text}</b>"
        widget.setText(text)
        widget.setObjectName(f"label_{name}")
        return widget

    def _get_tool_tip(
        self,
        attribute: Attribute[Any],
    ) -> Optional[str]:
        lines = []
        description = ""
        if attribute.description:
            description = attribute.description
        if attribute.deprecated:
            description = f"(Deprecated)\n{description}" if description else "(Deprecated)"
        if description:
            lines.append(description)
        if attribute.default is not None:
            lines.append(f"Default: {repr(attribute.default)}")
        return "\n".join(lines)

    def _create_field_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> QWidget:
        value_type = attribute.value_type
        widget: Optional[QWidget] = None
        for known_type, protocols in self._widget_types_protocols.items():
            if issubclass(value_type, known_type):
                widget = protocols.create(
                    name=name,
                    attribute=attribute,
                )
                break
        if not widget:
            widget = QLabel(self)
            widget.setWordWrap(True)
            widget.setTextFormat(Qt.TextFormat.RichText)
            widget.setStyleSheet("background-color: rgba(255, 0, 0, 50%);")
            widget.setText(
                f"No field for type <b>{value_type.__name__}</b>",
            )
        widget.setObjectName(f"field_{name}")

        return widget

    def _find_field_widget(
        self,
        name: str,
        widget_type: Type[T],
    ) -> T:
        return cast(T, self.findChild(widget_type, f"field_{name}"))

    def _update_attribute_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> None:
        value_type = attribute.value_type
        for known_type, protocols in self._widget_types_protocols.items():
            if issubclass(value_type, known_type):
                protocols.update(
                    name=name,
                    attribute=attribute,
                    value=value,
                )
                break

    def _create_attributes_container_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> AttributesContainerWidget:
        widget = AttributesContainerWidget(
            parent=self,
            container_class=attribute.value_type,
        )
        as_signal_instance(widget.container_changed).connect(
            partial(
                self._on_attributes_container_changed,
                widget,
                name,
                attribute,
            ),
        )
        return widget

    def _update_attributes_container_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        widget = self._find_field_widget(
            name=name,
            widget_type=AttributesContainerWidget,
        )
        widget.set_container(
            container=value,
        )
        return widget

    def _on_attributes_container_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        value: AttributesContainer,
    ) -> None:
        self._on_attribute_value_changed(
            widget=widget,
            name=name,
            attribute=attribute,
            value=value,
        )

    def _create_attributes_container_dict_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> AttributesContainerDictWidget:
        widget = AttributesContainerDictWidget(
            parent=self,
            container_dict_class=attribute.value_type,
        )
        as_signal_instance(widget.container_dict_changed).connect(
            partial(
                self._on_attributes_container_dict_changed,
                widget,
                name,
                attribute,
            ),
        )
        return widget

    def _update_attributes_container_dict_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        widget = self._find_field_widget(
            name=name,
            widget_type=AttributesContainerDictWidget,
        )
        widget.set_container_dict(
            container_dict=value,
        )
        return widget

    def _on_attributes_container_dict_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        value: AttributesContainerDict[AttributesContainer],
    ) -> None:
        self._on_attribute_value_changed(
            widget=widget,
            name=name,
            attribute=attribute,
            value=value,
        )

    def _create_attributes_container_list_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> AttributesContainerListWidget:
        widget = AttributesContainerListWidget(
            parent=self,
            container_list_class=attribute.value_type,
        )
        as_signal_instance(widget.container_list_changed).connect(
            partial(
                self._on_attributes_container_list_changed,
                widget,
                name,
                attribute,
            ),
        )
        return widget

    def _update_attributes_container_list_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        widget = self._find_field_widget(
            name=name,
            widget_type=AttributesContainerListWidget,
        )
        widget.set_container_list(
            container_list=value,
        )
        return widget

    def _on_attributes_container_list_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        value: AttributesContainerList[AttributesContainer],
    ) -> None:
        self._on_attribute_value_changed(
            widget=widget,
            name=name,
            attribute=attribute,
            value=value,
        )

    def _create_exportable_dict_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> ExportableDictWidget:
        widget = ExportableDictWidget(
            parent=self,
        )
        widget.set_model(
            model=ExportableDictModel(
                exportable_dict=attribute.value_type(),
            ),
        )
        as_signal_instance(widget.dataChanged).connect(
            partial(
                self._on_exportable_dict_changed,
                widget,
                name,
                attribute,
            ),
        )
        return widget

    def _update_exportable_dict_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        widget = self._find_field_widget(
            name=name,
            widget_type=ExportableDictWidget,
        )
        widget.set_model(
            model=ExportableDictModel(
                exportable_dict=value if value is not None else attribute.value_type(),
            ),
        )
        return widget

    def _on_exportable_dict_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        value: ExportableDict[Any, Any, Any, Any],
    ) -> None:
        self._on_attribute_value_changed(
            widget=widget,
            name=name,
            attribute=attribute,
            value=value,
        )

    def _create_exportable_list_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> ExportableListWidget:
        widget = ExportableListWidget(
            parent=self,
        )
        widget.set_model(
            model=ExportableListModel(
                exportable_list=attribute.value_type(),
            ),
        )
        as_signal_instance(widget.dataChanged).connect(
            partial(
                self._on_exportable_list_changed,
                widget,
                name,
                attribute,
            ),
        )
        return widget

    def _update_exportable_list_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        widget = self._find_field_widget(
            name=name,
            widget_type=ExportableListWidget,
        )
        widget.set_model(
            model=ExportableListModel(
                exportable_list=value if value is not None else attribute.value_type(),
            ),
        )
        return widget

    def _on_exportable_list_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        value: ExportableList[Any, Any],
    ) -> None:
        self._on_attribute_value_changed(
            widget=widget,
            name=name,
            attribute=attribute,
            value=value,
        )

    def _create_line_edit_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> QWidget:
        widget: Union[SecretLineEditWidget, QLineEdit]
        if attribute.secret:
            widget = SecretLineEditWidget(
                parent=self,
            )
            as_signal_instance(widget.text_edited).connect(
                partial(
                    self._on_line_edit_changed,
                    widget,
                    name,
                    attribute,
                ),
            )
        else:
            widget = QLineEdit(self)
            if isinstance(attribute.default, str):
                widget.setPlaceholderText(attribute.default)
            as_signal_instance(widget.textEdited).connect(
                partial(
                    self._on_line_edit_changed,
                    widget,
                    name,
                    attribute,
                ),
            )
        return widget

    def _update_line_edit_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        if attribute.secret:
            widget: Union[SecretLineEditWidget, QLineEdit] = self._find_field_widget(
                name=name,
                widget_type=SecretLineEditWidget,
            )
        else:
            widget = self._find_field_widget(
                name=name,
                widget_type=QLineEdit,
            )
        if value != attribute.default:
            widget.setText(value)
        return widget

    def _on_line_edit_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        value: str,
    ) -> None:
        clean_value: Optional[str] = value
        if not clean_value:
            clean_value = None
        self._on_attribute_value_changed(
            widget=widget,
            name=name,
            attribute=attribute,
            value=clean_value,
        )

    def _create_check_box_widget(
        self,
        name: str,
        attribute: Attribute[Any],
    ) -> HintedCheckBoxWidget:
        widget = HintedCheckBoxWidget(self)
        widget.setTristate(True)
        default_hint = "Yes" if attribute.default else "No"
        widget.set_state_hint(Qt.CheckState.PartiallyChecked, f"Default: {default_hint}")
        widget.setCheckState(Qt.CheckState.PartiallyChecked)
        as_signal_instance(widget.stateChanged).connect(
            partial(
                self._on_check_box_changed,
                widget,
                name,
                attribute,
            ),
        )
        return widget

    def _update_check_box_widget(
        self,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> QWidget:
        state = Qt.CheckState.Checked if value else Qt.CheckState.Unchecked
        if value == attribute.default:
            state = Qt.CheckState.PartiallyChecked
        widget = self._find_field_widget(
            name=name,
            widget_type=HintedCheckBoxWidget,
        )
        widget.setCheckState(state)
        return widget

    def _on_check_box_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        state: Qt.CheckState,
    ) -> None:
        value = None
        if state != Qt.CheckState.PartiallyChecked:
            value = state == Qt.CheckState.Checked
        self._on_attribute_value_changed(
            widget=widget,
            name=name,
            attribute=attribute,
            value=value,
        )

    def _on_attribute_value_changed(
        self,
        widget: QWidget,
        name: str,
        attribute: Attribute[Any],
        value: Any,
    ) -> None:
        if self._container is None:
            self._container = self._container_class()
        setattr(self._container, name, value)
        as_signal_instance(self.container_changed).emit(self._container)



"""Models and functions used for cryptographic keys."""
from __future__ import annotations

from typing import Generator

from ywh2bt.core.crypt.error import CryptError


class Key:
    """A simple crypt key."""

    _key_bytes: bytes

    def __init__(
        self,
        key_bytes: bytes,
    ):
        """
        Initialize self.

        Args:
            key_bytes: key in bytes

        Raises:
            CryptError: if an error occurred
        """
        if not isinstance(key_bytes, bytes):
            raise CryptError(f"Key should be {bytes}. {type(key_bytes)} given.")
        self._key_bytes = key_bytes

    def generator(
        self,
    ) -> Generator[int, None, None]:
        """
        Get a generator that yields each byte of the key.

        Yields:
            each byte of the key
        """
        idx = 0
        key_size = len(self._key_bytes)
        while True:
            yield self._key_bytes[idx % key_size]
            idx += 1

    def __str__(
        self,
    ) -> str:
        """
        Get a string representation of the key.

        Returns:
            The string representation of the key.
        """
        return f"Key(size={len(self._key_bytes)})"

    @classmethod
    def from_str(
        cls,
        key: str,
    ) -> Key:
        """
        Build a key from a string.

        Args:
            key: a key in a string format.

        Returns:
            The key.
        """
        return Key(str.encode(key))



"""Models used for the configuration of a GitHub tracker."""
from typing import (
    Any,
    Dict,
    Optional,
    Text,
)

from ywh2bt.core.configuration.attribute import (
    Attribute,
    BoolAttributeType,
    StrAttributeType,
)
from ywh2bt.core.configuration.error import (
    AttributesError,
    MissingAttributeError,
)
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.validator import (
    not_blank_validator,
    url_validator,
)


class GitHubConfiguration(TrackerConfiguration):
    """A class describing the configuration of a GitHub tracker."""

    url: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="API URL",
        description="Base URL of the GitHub API",
        default="https://api.github.com",
        validator=url_validator,
    )
    token: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="API token",
        description="User private token for the GitHub API",
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    project: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Project path",
        description="Path to the project on GitHub",
        required=True,
        validator=not_blank_validator,
    )
    verify: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Verify TLS",
        description="Verify server's TLS certificate",
        default=True,
    )
    github_cdn_on: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Use CDN",
        description="Enable or disable saving attachment file to GitHub CDN",
        default=False,
    )
    login: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Login",
        description="User login for the GitHub server. Required only if github_cdn_on is true",
        secret=False,
    )
    password: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Password",
        description="User password for the GitHub server. Required only if github_cdn_on is true",
        secret=True,
    )

    def __init__(
        self,
        url: Optional[Text] = None,
        token: Optional[Text] = None,
        project: Optional[Text] = None,
        verify: Optional[bool] = None,
        github_cdn_on: Optional[bool] = None,
        login: Optional[Text] = None,
        password: Optional[Text] = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            url: a GitHub API URL
            token: a GitHub API token
            project: a GitHub project name
            verify: a flag indicating whether to check SSL/TLS connection
            github_cdn_on: a flag indicating whether to save attachment file to GitHub CDN
            login: a GitLab user login ; used when github_cdn_on is true
            password: a GitLab user password ; used when github_cdn_on is true
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.url = url
        self.token = token
        self.project = project
        self.verify = verify
        self.github_cdn_on = github_cdn_on
        self.login = login
        self.password = password

    def validate(self) -> None:
        """
        Validate the container.

        # noqa: DAR401, DAR402

        Raises:
            AttributesError: if the container is not valid
        """
        super_error: Optional[AttributesError] = None
        try:
            super().validate()
        except AttributesError as e:
            super_error = e

        errors = {}
        if self.github_cdn_on:
            errors.update(
                self._ensure_login_password(),
            )

        if errors:
            if not super_error:
                super_error = AttributesError(
                    message="Validation error",
                    errors={},
                    context=self,
                )
            super_error.errors.update(errors)

        if super_error:
            raise super_error

    def _ensure_login_password(
        self,
    ) -> Dict[str, MissingAttributeError]:
        errors = {}
        if self.login is None:
            errors["login"] = MissingAttributeError(
                message="Expecting value for attribute 'login' when 'github_on_cdn' is True",
                context=self,
            )
        if self.password is None:
            errors["password"] = MissingAttributeError(
                message="Expecting value for attribute 'password' when 'github_on_cdn' is True",
                context=self,
            )
        return errors


TrackerConfiguration.register_subtype(
    subtype_name="github",
    subtype_class=GitHubConfiguration,
)



"""Models and functions used in data typing."""
from typing import (
    Any,
    Union,
    cast,
)

from PySide6.QtCore import (
    Signal,
    SignalInstance,
)


def as_signal_instance(
    signal: Union[Signal, Any],
) -> SignalInstance:
    """
    Cast a Signal object to a SignalInstance.

    PySide6 magic / static typing / mypy.

    Args:
        signal: a signal object

    Returns:
        The signal instance
    """
    return cast(SignalInstance, signal)



"""Models and functions for CLI test."""
import argparse

from ywh2bt.cli.error import CliError
from ywh2bt.cli.listener import CliTesterListener
from ywh2bt.core.configuration.error import BaseAttributeError
from ywh2bt.core.core import load_configuration
from ywh2bt.core.factories.tracker_clients import TrackerClientsFactory
from ywh2bt.core.factories.yeswehack_api_clients import YesWeHackApiClientsFactory
from ywh2bt.core.tester.error import TesterError
from ywh2bt.core.tester.tester import Tester


def test(
    args: argparse.Namespace,
) -> None:
    """
    Test YWH and trackers clients.

    Args:
        args: command line arguments

    Raises:
        CliError: if an error occurred
    """
    configuration = load_configuration(
        config_format=args.config_format,
        config_file=args.config_file,
    )
    try:
        configuration.validate()
    except BaseAttributeError as validation_error:
        raise CliError("Invalid configuration") from validation_error
    tester = Tester(
        configuration=configuration,
        yes_we_hack_api_clients_factory=YesWeHackApiClientsFactory(),
        tracker_clients_factory=TrackerClientsFactory(),
        listener=CliTesterListener(),
    )
    try:
        tester.test()
    except TesterError as testing_error:
        raise CliError("Testing error") from testing_error



"""Models and functions used for resources updates."""
import os
import sys
from pathlib import Path
from string import Template
from typing import Generator


resources_file_template_contents = """<!DOCTYPE RCC>
<RCC version="1.0">
    <qresource>
        ${glued_lines}
    </qresource>
</RCC>
"""
resources_file_template = Template(resources_file_template_contents)

resources_file_line_template = Template(
    "<file>${path}</file>",
)
resources_file_line_glue = "\n        "


def update() -> None:
    """Create or update the contents of resources.qrc based on the directory listing of ./resources/."""
    base_dir = os.path.dirname(__file__)
    content = _make_file_content(
        base_dir=base_dir,
    )
    file_path = f"{base_dir}/resources.qrc"
    with open(file_path, "w") as f:
        f.write(content)
    sys.stdout.write(f"{file_path} updated.")
    sys.stdout.write("\n")


def _make_file_content(
    base_dir: str,
) -> str:
    resources_dir = "resources"
    resources = _find_resources(
        base_dir=os.path.join(
            base_dir,
            resources_dir,
        ),
    )
    lines = map(
        lambda path: resources_file_line_template.substitute(
            path=f"{resources_dir}/{path}",
        ),
        resources,
    )
    return resources_file_template.substitute(
        glued_lines=resources_file_line_glue.join(lines),
    )


def _find_resources(
    base_dir: str,
) -> Generator[str, None, None]:
    for path in Path(base_dir).rglob("*"):
        if not path.is_file():
            continue
        yield str(path.relative_to(base_dir))


if __name__ == "__main__":
    update()



"""Models and classes related to exportable objects."""
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Generic,
    TypeVar,
)


T = TypeVar("T")


class Exportable(Generic[T], ABC):
    """A class representing an object that can be exported."""

    @abstractmethod
    def export(self) -> T:
        """Export the object."""



"""Models used for the configuration of YesWeHack."""
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    cast,
)

from ywh2bt.core.configuration.attribute import (
    Attribute,
    AttributesContainer,
    AttributesContainerDict,
    AttributesContainerList,
    BoolAttributeType,
    ExportableList,
    StrAttributeType,
)
from ywh2bt.core.configuration.validator import (
    length_one_validator,
    not_blank_validator,
    not_empty_validator,
    url_validator,
)


class Bugtrackers(ExportableList[str, str]):
    """A list of bugtrackers."""


BugtrackersNameAttributeType = Union[
    Optional[Bugtrackers],
    Attribute[Bugtrackers],
]
BugtrackersNameInitType = Union[
    Optional[Bugtrackers],
    Optional[List[str]],
]


class SynchronizeOptions(AttributesContainer):
    """Synchronization option."""

    upload_private_comments: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload private comments",
        description="Upload the report private comments into the bug tracker",
        default=False,
    )
    upload_assign_comments: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload assignments comments",
        description="Upload the comments of assignments into the bug tracker",
        default=False,
    )
    upload_public_comments: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload public comments",
        description="Upload the report public comments into the bug tracker",
        default=False,
    )
    upload_cvss_updates: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload CVSS updates",
        description="Upload the report CVSS updates into the bug tracker",
        default=False,
    )
    upload_details_updates: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload details updates",
        description="Upload the report details updates into the bug tracker",
        default=False,
    )
    upload_priority_updates: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload priority updates",
        description="Upload the report priority updates into the bug tracker",
        default=False,
    )
    upload_rewards: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload rewards",
        description="Upload the report rewards into the bug tracker",
        default=False,
    )
    upload_status_updates: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload status updates",
        description="Upload the report status updates into the bug tracker",
        default=False,
    )
    upload_transfer_updates: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Upload transfer updates",
        description="Upload the report transfer updates into the bug tracker",
        default=False,
    )
    recreate_missing_issues: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Recreate missing issues",
        description=(
            "Recreate issues that were created by a previous synchronization "
            + "but are not found into the bug tracker anymore"
        ),
        default=True,
    )

    def __init__(
        self,
        upload_private_comments: Optional[bool] = None,
        upload_public_comments: Optional[bool] = None,
        upload_assign_comments: Optional[bool] = None,
        upload_details_updates: Optional[bool] = None,
        upload_cvss_updates: Optional[bool] = None,
        upload_priority_updates: Optional[bool] = None,
        upload_rewards: Optional[bool] = None,
        upload_status_updates: Optional[bool] = None,
        upload_transfer_updates: Optional[bool] = None,
        recreate_missing_issues: Optional[bool] = None,
        **kwargs: Any,
    ):
        """
        Initialize self.

        Args:
            upload_private_comments: a flag indicating whether to upload private comments to the bug tracker
            upload_assign_comments: a flag indicating whether to upload assign comments to the bug tracker
            upload_public_comments: a flag indicating whether to upload public comments to the bug tracker
            upload_cvss_updates: a flag indicating whether to upload CVSS updates to the bug tracker
            upload_details_updates: a flag indicating whether to upload details updates to the bug tracker
            upload_priority_updates: a flag indicating whether to upload priority updates into the bug tracker
            upload_rewards: a flag indicating whether to upload rewards to the bug tracker
            upload_status_updates: a flag indicating whether to upload status updates to the bug tracker
            upload_transfer_updates: a flag indicating whether to upload transfer updates to the bug tracker
            recreate_missing_issues: a flag indicating whether to recreate missing issues
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.upload_private_comments = upload_private_comments
        self.upload_assign_comments = upload_assign_comments
        self.upload_public_comments = upload_public_comments
        self.upload_cvss_updates = upload_cvss_updates
        self.upload_details_updates = upload_details_updates
        self.upload_priority_updates = upload_priority_updates
        self.upload_rewards = upload_rewards
        self.upload_status_updates = upload_status_updates
        self.upload_transfer_updates = upload_transfer_updates
        self.recreate_missing_issues = recreate_missing_issues


SynchronizeOptionsAttributeType = Union[
    Dict[str, bool],
    SynchronizeOptions,
    Attribute[SynchronizeOptions],
]
SynchronizeOptionsInitType = Union[
    Optional[SynchronizeOptions],
    Optional[Dict[str, bool]],
]


class FeedbackOptions(AttributesContainer):
    """Feedback option."""

    download_tracker_comments: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Download bug trackers comments",
        description="Download comments from the bug tracker and put them into the report",
        default=False,
    )
    issue_closed_to_report_afv: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Issue closed to report AFV",
        description='Set the report status to "Ask for Fix Verification" when the tracker issue is closed',
        default=False,
    )

    def __init__(
        self,
        download_tracker_comments: Optional[bool] = None,
        issue_closed_to_report_afv: Optional[bool] = None,
        **kwargs: Any,
    ):
        """
        Initialize self.

        Args:
            download_tracker_comments:
            a flag indicating whether to download comments from the bug tracker and put them into the report
            issue_closed_to_report_afv:
            a flag indicating whether to set report status to "Ask for Fix Verification" when tracker issue is closed
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.download_tracker_comments = download_tracker_comments
        self.issue_closed_to_report_afv = issue_closed_to_report_afv


FeedbackOptionsAttributeType = Union[
    Dict[str, bool],
    FeedbackOptions,
    Attribute[FeedbackOptions],
]
FeedbackOptionsInitType = Union[
    Optional[FeedbackOptions],
    Optional[Dict[str, bool]],
]


class Program(AttributesContainer):
    """A program and its associated bugtrackers."""

    slug: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Program slug",
        description="Program slug",
        required=True,
        validator=not_blank_validator,
    )
    synchronize_options: SynchronizeOptionsAttributeType = Attribute.create(
        value_type=SynchronizeOptions,
        short_description="Synchronization options",
        required=True,
    )
    feedback_options: FeedbackOptionsAttributeType = Attribute.create(
        value_type=FeedbackOptions,
        short_description="Feedback options",
        required=True,
    )
    bugtrackers_name: BugtrackersNameAttributeType = Attribute.create(
        value_type=Bugtrackers,
        short_description="Bug trackers",
        description="Bug trackers names",
        required=True,
        validator=length_one_validator,
    )

    def __init__(
        self,
        slug: Optional[str] = None,
        synchronize_options: SynchronizeOptionsInitType = None,
        feedback_options: FeedbackOptionsInitType = None,
        bugtrackers_name: BugtrackersNameInitType = None,
        **kwargs: Any,
    ):
        """
        Initialize the program.

        Args:
            slug: a program slug
            synchronize_options: synchronization options
            feedback_options: feedback options
            bugtrackers_name: a list of bugtrackers
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.slug = slug
        if not synchronize_options:
            synchronize_options = SynchronizeOptions()
        self.synchronize_options = synchronize_options
        if not feedback_options:
            feedback_options = FeedbackOptions()
        self.feedback_options = feedback_options
        if bugtrackers_name:
            if isinstance(bugtrackers_name, List):
                self.bugtrackers_name = Bugtrackers(bugtrackers_name)
            else:
                self.bugtrackers_name = cast(Bugtrackers, bugtrackers_name)


class Programs(AttributesContainerList[Program]):
    """A list of programs."""

    def __init__(
        self,
        items: Optional[List[Any]] = None,
    ) -> None:
        """
        Initialize the list of programs.

        Args:
            items: a list of programs
        """
        super().__init__(
            values_type=Program,
            items=items,
        )


ProgramsAttributeType = Union[
    List[Union[Program, Dict[str, Any]]],
    Optional[Programs],
    Attribute[Programs],
]
ProgramsInitType = Union[
    Optional[Programs],
    Optional[List[Union[Program, Dict[str, Any]]]],
]


class YesWeHackConfiguration(AttributesContainer):
    """A configuration for YesWeHack."""

    api_url: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="API URL",
        description="Base URL of the YWH API",
        default="https://api.yeswehack.com",
        validator=url_validator,
    )
    pat: StrAttributeType = Attribute.create(
        value_type=str,
        short_description="Personal Access Token",
        description="Personal Access Token",
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    verify: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description="Verify TLS",
        description="Verify server's TLS certificate",
        default=True,
    )
    programs: ProgramsAttributeType = Attribute.create(
        value_type=Programs,
        short_description="Programs",
        description="Programs",
        required=True,
        validator=not_empty_validator,
    )

    def __init__(
        self,
        api_url: Optional[str] = None,
        pat: Optional[str] = None,
        verify: Optional[bool] = None,
        programs: ProgramsInitType = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            api_url: a YesWeHack API URL
            pat: a Personal Access Token
            verify: a flag indicating whether to check SSL/TLS connection
            programs: a list of Programs
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.api_url = api_url
        self.pat = pat
        self.verify = verify
        self.programs = programs


class YesWeHackConfigurations(AttributesContainerDict[YesWeHackConfiguration]):
    """A list of YesWeHack configurations."""

    def __init__(
        self,
        **kwargs: Any,
    ):
        """
        Initialize the list of configurations.

        Args:
            kwargs: a list of configurations
        """
        super().__init__(
            values_type=YesWeHackConfiguration,
            items=kwargs,
        )



"""Models and functions used for formatting reports data."""

from string import Template

from ywh2bt.core.api.formatter.formatter import ReportMessageFormatter
from ywh2bt.core.html import ywh_html_to_markdown


REPORT_TITLE_TEMPLATE = "${local_id} : ${title}"
REPORT_DESCRIPTION_TEMPLATE = """Title: ${local_id} : ${title}

Priority: ${priority_name}
${bug_type_label}: ${bug_type_name} (${bug_type_link})
${bug_type_label} remediation: ${bug_type_remediation_link}
${scope_label}: ${scope}
Severity: ${cvss_criticity}
    - Score: ${cvss_score}
    - Vector: ${cvss_vector}
${end_point_label}: ${end_point}
${vulnerable_part_label}: ${vulnerable_part}
${part_name_label}: ${part_name}
${payload_sample_label}: ${payload_sample}
Technical Environment: ${technical_environment}

${description}
"""
CLOSE_LOG_TEMPLATE = """
Report closed:
${old_status} -> ${new_status}

Comment:
${comment}
"""
COMMENT_BODY_TEMPLATE = """
Log: #${log_id}
Date: ${date}

${body}
"""
COMMENT_LOG_TEMPLATE = """
Comment:
${comment}
"""
CVSS_UPDATE_LOG_TEMPLATE = """
CVSS updated:
Severity:
    - Old value: ${old_cvss_criticity}
    - New Value: ${new_cvss_criticity}
Score:
    - Old value: ${old_cvss_score}
    - New Value: ${new_cvss_score}
"""
STATUS_UPDATE_LOG_TEMPLATE = """
Status updated:
${old_status} -> ${new_status}

Comment:
${comment}
"""
FIX_VERIFIED_LOG_TEMPLATE = """
Fix verification:
${verified}

Comment:
${comment}
"""
DETAILS_UPDATE_LOG_TEMPLATE = """
Details updated:
${details_lines}
"""
DETAILS_UPDATE_LOG_LINE_TEMPLATE = """${updated_property}:
    - Old value: ${old_value}
    - New Value: ${new_value}
"""
PRIORITY_UPDATE_LOG_TEMPLATE = """
Priority updated:
${new_priority}
"""
REWARD_LOG_TEMPLATE = """
Comment from reward:
${comment}
"""
TRANSFER_LOG_TEMPLATE = """
Report transferred:
    - ${old_program}
    - ${program}
"""
ASK_FOR_FIX_VERIFICATION_STATUS_LOG_TEMPLATE = """
Fix verification:
${new_ask_for_fix_verification_status}

Comment:
${comment}
"""


class ServiceNowReportMessageFormatter(ReportMessageFormatter):
    """A report formatter to ServiceNow."""

    def __init__(self) -> None:
        """Initialize self."""
        super().__init__(
            report_title_template=Template(REPORT_TITLE_TEMPLATE),
            report_description_template=Template(REPORT_DESCRIPTION_TEMPLATE),
            close_log_template=Template(CLOSE_LOG_TEMPLATE),
            comment_body_template=Template(COMMENT_BODY_TEMPLATE),
            comment_log_template=Template(COMMENT_LOG_TEMPLATE),
            cvss_update_log_template=Template(CVSS_UPDATE_LOG_TEMPLATE),
            status_update_log_template=Template(STATUS_UPDATE_LOG_TEMPLATE),
            fix_verified_log_template=Template(FIX_VERIFIED_LOG_TEMPLATE),
            details_update_log_template=Template(DETAILS_UPDATE_LOG_TEMPLATE),
            details_update_log_line_template=Template(DETAILS_UPDATE_LOG_LINE_TEMPLATE),
            priority_update_log_template=Template(PRIORITY_UPDATE_LOG_TEMPLATE),
            reward_log_template=Template(REWARD_LOG_TEMPLATE),
            transfer_log_template=Template(TRANSFER_LOG_TEMPLATE),
            ask_for_fix_verification_status_log_template=Template(ASK_FOR_FIX_VERIFICATION_STATUS_LOG_TEMPLATE),
        )

    def transform_report_description_html(
        self,
        description_html: str,
    ) -> str:
        """
        Transform the report description.

        Args:
            description_html: a report description in HTML

        Returns:
            a transformed description
        """
        return ywh_html_to_markdown(
            html=description_html,
        )

    def transform_html(
        self,
        html: str,
    ) -> str:
        """
        Transform from HTML.

        Args:
            html: HTML

        Returns:
            HTML transformed to markdown
        """
        return ywh_html_to_markdown(
            html=html,
        )



"""Models and functions used for main GUI window."""
from __future__ import annotations

import sys
from typing import (
    List,
    Optional,
    Union,
    cast,
)

from PySide6.QtCore import (
    QByteArray,
    QSettings,
    QSize,
    Qt,
    Signal,
)
from PySide6.QtGui import (
    QAction,
    QCloseEvent,
    QIcon,
    QKeySequence,
)
from PySide6.QtWidgets import (
    QMainWindow,
    QMenu,
    QMessageBox,
    QSizePolicy,
    QStyle,
    QToolBar,
    QWidget,
)

from ywh2bt.core.core import AVAILABLE_FORMATS
from ywh2bt.gui.dialog.file import FileFormatDialogFilters
from ywh2bt.gui.dialog.schema import SchemaDocumentationDialog
from ywh2bt.gui.widgets.logs_widget import LogEntry
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.root_configurations_widget import RootConfigurationsWidget
from ywh2bt.gui.widgets.typing import as_signal_instance
from ywh2bt.version import __VERSION__


__UI_VERSION__ = 1

_MENU_SEPARATOR = "-"
_TOOL_BAR_SEPARATOR = "-"
_TOOL_BAR_SPACER = "<->"


class MainWindow(QMainWindow):
    """Main window."""

    _log_entry_available: Signal = Signal(LogEntry)

    _file_format_dialog_filters: FileFormatDialogFilters
    _settings: QSettings
    _schema_doc_dialog: Optional[SchemaDocumentationDialog]

    _action_new: QAction
    _action_open: QAction
    _action_save: QAction
    _action_save_as: QAction
    _action_reload: QAction
    _action_test: QAction
    _action_sync: QAction
    _action_exit: QAction
    _action_schema_doc: QAction
    _action_about: QAction

    _root_configurations_widget: RootConfigurationsWidget

    def __init__(
        self,
        settings: QSettings,
    ) -> None:
        """
        Initialize self.

        Args:
            settings: application settings
        """
        super().__init__()
        self._file_format_dialog_filters = FileFormatDialogFilters(
            formats=AVAILABLE_FORMATS,
        )
        self._settings = settings
        self._schema_doc_dialog = None
        self._init_ui()
        self._restore_settings()

    def _init_ui(self) -> None:
        self._create_actions()
        self._create_menus()

        tool_bar = self._create_main_tool_bar()
        self.addToolBar(
            Qt.ToolBarArea.LeftToolBarArea,
            tool_bar,
        )

        self._root_configurations_widget = self._create_root_configurations_widget()

        self.setCentralWidget(self._root_configurations_widget)

        self.statusBar().show()

    def _create_actions(
        self,
    ) -> None:
        self._action_new = self._create_new_action()
        self._action_open = self._create_open_action()
        self._action_save = self._create_save_action()
        self._action_save_as = self._create_save_as_action()
        self._action_reload = self._create_reload_action()
        self._action_test = self._create_test_action()
        self._action_sync = self._create_sync_action()
        self._action_exit = self._create_exit_action()
        self._action_schema_doc = self._create_schema_doc_action()
        self._action_about = self._create_about_action()

    def _create_menus(
        self,
    ) -> None:
        bar = self.menuBar()

        file_menu = bar.addMenu("&File")
        file_menu.setObjectName("menu_file")
        _add_menu_items(
            menu=file_menu,
            items=[
                self._action_new,
                self._action_open,
                self._action_save,
                self._action_save_as,
                self._action_reload,
                _MENU_SEPARATOR,
                self._action_exit,
            ],
        )

        run_menu = bar.addMenu("&Run")
        run_menu.setObjectName("menu_run")
        _add_menu_items(
            menu=run_menu,
            items=[
                self._action_test,
                self._action_sync,
            ],
        )

        help_menu = bar.addMenu("&Help")
        help_menu.setObjectName("menu_help")
        _add_menu_items(
            menu=help_menu,
            items=[
                self._action_schema_doc,
                self._action_about,
            ],
        )

    def _create_new_action(
        self,
    ) -> QAction:
        action = QAction("&New", self)
        action.setObjectName("action_new")
        action.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_FileIcon))
        action.setShortcuts(QKeySequence.StandardKey.New)
        action.setStatusTip("Create a new configuration file")
        as_signal_instance(action.triggered).connect(
            self._on_new_file_triggered,
        )
        return action

    def _create_open_action(
        self,
    ) -> QAction:
        action = QAction("&Open...", self)
        action.setObjectName("action_open")
        action.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_DialogOpenButton))
        action.setShortcuts(QKeySequence.StandardKey.Open)
        action.setStatusTip("Open a configuration file")
        as_signal_instance(action.triggered).connect(
            self._on_open_file_triggered,
        )
        return action

    def _create_save_action(
        self,
    ) -> QAction:
        action = QAction("&Save...", self)
        action.setObjectName("action_save")
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_DialogSaveButton))
        action.setShortcuts(QKeySequence.StandardKey.Save)
        action.setStatusTip("Save the current configuration file")
        as_signal_instance(action.triggered).connect(
            self._on_save_configuration_triggered,
        )
        return action

    def _create_save_as_action(
        self,
    ) -> QAction:
        action = QAction("&Save as...", self)
        action.setObjectName("action_save_as")
        action.setEnabled(False)
        action.setStatusTip("Save the current configuration in a new file")
        as_signal_instance(action.triggered).connect(
            self._on_save_as_configuration_triggered,
        )
        return action

    def _create_reload_action(
        self,
    ) -> QAction:
        action = QAction("&Reload from disk", self)
        action.setObjectName("action_reload")
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_BrowserReload))
        action.setShortcuts(QKeySequence.StandardKey.Refresh)
        action.setStatusTip("Reload the current configuration file from the disk")
        as_signal_instance(action.triggered).connect(
            self._on_reload_configuration_triggered,
        )
        return action

    def _create_test_action(
        self,
    ) -> QAction:
        action = QAction("&Test...", self)
        action.setObjectName("action_test")
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_DialogApplyButton))
        action.setStatusTip("Test the connection to the trackers and to YesWeHack programs")
        as_signal_instance(action.triggered).connect(
            self._on_test_configuration_triggered,
        )
        return action

    def _create_sync_action(
        self,
    ) -> QAction:
        action = QAction("&Sync...", self)
        action.setObjectName("action_sync")
        action.setEnabled(False)
        action.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay))
        action.setStatusTip("Synchronize the current configuration file")
        as_signal_instance(action.triggered).connect(
            self._on_sync_configuration_triggered,
        )
        return action

    def _create_about_action(
        self,
    ) -> QAction:
        action = QAction("&About", self)
        action.setObjectName("action_about")
        action.setIcon(QIcon(":/resources/icons/ywh2bt.png"))
        action.setStatusTip("Show information about this application")
        as_signal_instance(action.triggered).connect(
            self._on_about_triggered,
        )
        return action

    def _create_schema_doc_action(
        self,
    ) -> QAction:
        action = QAction("&Schema documentation", self)
        action.setObjectName("action_schema_doc")
        action.setIcon(self.style().standardIcon(QStyle.StandardPixmap.SP_DialogHelpButton))
        action.setStatusTip("Show configuration schema documentation")
        as_signal_instance(action.triggered).connect(
            self._on_schema_doc_triggered,
        )
        return action

    def _create_exit_action(
        self,
    ) -> QAction:
        action = QAction("E&xit", self)
        action.setObjectName("action_exit")
        action.setShortcuts(QKeySequence.StandardKey.Quit)
        action.setStatusTip("Exit the application")
        as_signal_instance(action.triggered).connect(
            self._on_exit_triggered,
        )
        return action

    def _create_main_tool_bar(
        self,
    ) -> QToolBar:
        tool_bar = self.addToolBar("Main tools")
        tool_bar.setObjectName("tools_tool_bar")

        _add_tool_bar_items(
            tool_bar=tool_bar,
            items=[
                self._action_new,
                self._action_open,
                self._action_save,
                self._action_reload,
                _TOOL_BAR_SEPARATOR,
                self._action_test,
                self._action_sync,
                _TOOL_BAR_SPACER,
                _TOOL_BAR_SEPARATOR,
                self._action_schema_doc,
                self._action_about,
            ],
        )

        return tool_bar

    def _create_empty_widget(
        self,
    ) -> QWidget:
        toolbar = QToolBar(self)
        toolbar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)

        spacer_begin = QWidget(toolbar)
        spacer_begin.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        spacer_end = QWidget(toolbar)
        spacer_end.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)

        toolbar.addWidget(spacer_begin)
        toolbar.addAction(self._action_new)
        toolbar.addAction(self._action_open)
        toolbar.addWidget(spacer_end)

        return toolbar

    def _create_root_configurations_widget(
        self,
    ) -> RootConfigurationsWidget:
        widget = RootConfigurationsWidget(
            parent=self,
            settings=self._settings,
            empty_widget=self._create_empty_widget(),
        )
        as_signal_instance(widget.entry_modified).connect(
            self._on_entry_modified,
        )
        as_signal_instance(widget.current_entry_changed).connect(
            self._on_current_entry_changed,
        )
        as_signal_instance(widget.entry_saved).connect(
            self._on_entry_saved,
        )
        return widget

    def _on_entry_modified(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        self._update_actions(
            entry=entry,
        )

    def _on_entry_saved(
        self,
        entry: RootConfigurationEntry,
    ) -> None:
        self.statusBar().showMessage(
            f"{entry.name} saved",
        )

    def _on_current_entry_changed(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        self._update_actions(
            entry=entry,
        )

    def _update_actions(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        self._update_save_action(
            entry=entry,
        )
        self._update_save_as_action(
            entry=entry,
        )
        self._update_reload_action(
            entry=entry,
        )
        self._update_test_action(
            entry=entry,
        )
        self._update_sync_action(
            entry=entry,
        )

    def _update_save_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and not entry.is_empty() and entry.has_changed():
            enabled = True
        self._action_save.setEnabled(enabled)

    def _update_save_as_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and not entry.is_empty():
            enabled = True
        self._action_save_as.setEnabled(enabled)

    def _update_reload_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and entry.file:
            enabled = True
        self._action_reload.setEnabled(enabled)

    def _update_test_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and entry.configuration and entry.configuration.is_valid():
            enabled = True
        self._action_test.setEnabled(enabled)

    def _update_sync_action(
        self,
        entry: Optional[RootConfigurationEntry],
    ) -> None:
        enabled = False
        if entry and entry.configuration and entry.configuration.is_valid():
            enabled = True
        self._action_sync.setEnabled(enabled)

    def _on_new_file_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.create_new_configuration()

    def _on_open_file_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.prompt_open_configuration()

    def _on_save_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.save_current_entry()

    def _on_save_as_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.save_as_current_entry()

    def _on_reload_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.reload_current_configuration()

    def _on_test_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.test_current_entry()

    def _on_sync_configuration_triggered(
        self,
    ) -> None:
        self._root_configurations_widget.synchronize_current_entry()

    def _on_exit_triggered(
        self,
    ) -> None:
        self.close()

    def _on_about_triggered(
        self,
    ) -> None:
        python_version = ".".join(map(str, sys.version_info[:3]))
        about_text = [
            (
                "<b><a href='https://github.com/yeswehack/ywh2bugtracker'>ywh2bt</a> "
                "by <a href='https://yeswehack.com'>YesWeHack</a></b>"
            ),
            f"<b>Core version</b>: {__VERSION__}",
            f"<b>UI version</b>: {__UI_VERSION__}",
            f"<b>Python</b>: {python_version}",
        ]
        QMessageBox.about(
            self,
            "ywh2bt - About",
            "<br/>\n".join(about_text),
        )

    def _on_schema_doc_triggered(
        self,
    ) -> None:
        if not self._schema_doc_dialog:
            self._schema_doc_dialog = SchemaDocumentationDialog(
                self,
            )
            self._schema_doc_dialog.resize(
                QSize(600, 600),
            )
            self._restore_schema_doc_geometry_from_settings()
        self._schema_doc_dialog.show()

    def open_file(
        self,
        file_path: str,
        file_format: str,
    ) -> None:
        self._root_configurations_widget.try_open_configuration_file(
            file_path=file_path,
            file_format=file_format,
        )

    def closeEvent(
        self,
        event: QCloseEvent,
    ) -> None:
        """
        Handle 'close' event.

        Args:
            event: a close event
        """
        if self._root_configurations_widget.has_changed_entries():
            reply = QMessageBox.question(
                self,
                "Exit",
                "Are you sure you want to exit?\nUnsaved changes might be lost.",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if reply == QMessageBox.StandardButton.No:
                event.ignore()
                return
        self._save_settings()
        event.accept()
        super().closeEvent(event)

    def _save_settings(
        self,
    ) -> None:
        self._settings.setValue("main_geometry", self.saveGeometry().toBase64())
        self._settings.setValue("main_state", self.saveState(__UI_VERSION__).toBase64())
        if self._schema_doc_dialog:
            self._settings.setValue("schema_doc_geometry", self._schema_doc_dialog.saveGeometry().toBase64())

    def _restore_settings(
        self,
    ) -> None:
        self._restore_geometry_from_settings()
        self._restore_state_from_settings()

    def _restore_geometry_from_settings(
        self,
    ) -> None:
        if not self._settings.contains("main_geometry"):
            return
        geometry = QByteArray.fromBase64(cast(QByteArray, self._settings.value("main_geometry")))
        if geometry:
            self.restoreGeometry(geometry)

    def _restore_state_from_settings(
        self,
    ) -> None:
        if not self._settings.contains("main_state"):
            return
        state = QByteArray.fromBase64(cast(QByteArray, self._settings.value("main_state")))
        if state:
            self.restoreState(state, __UI_VERSION__)

    def _restore_schema_doc_geometry_from_settings(
        self,
    ) -> None:
        if not self._schema_doc_dialog or not self._settings.contains("schema_doc_geometry"):
            return
        geometry = QByteArray.fromBase64(cast(QByteArray, self._settings.value("schema_doc_geometry")))
        if geometry:
            self._schema_doc_dialog.restoreGeometry(geometry)


def _add_menu_items(
    menu: QMenu,
    items: List[Union[str, QAction]],
) -> None:
    for item in items:
        if item == _MENU_SEPARATOR:
            menu.addSeparator()
        elif isinstance(item, QAction):
            menu.addAction(item)


def _add_tool_bar_items(
    tool_bar: QToolBar,
    items: List[Union[str, QAction]],
) -> None:
    for item in items:
        if item == _TOOL_BAR_SEPARATOR:
            tool_bar.addSeparator()
        elif item == _TOOL_BAR_SPACER:
            spacer = QWidget(tool_bar)
            spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
            tool_bar.addWidget(spacer)
        elif isinstance(item, QAction):
            tool_bar.addAction(item)



"""Models and functions used in error dialogs."""
from PySide6.QtWidgets import (
    QMessageBox,
    QWidget,
)

from ywh2bt.core.error import error_to_string


class ErrorDialogMixin:
    """Mixin for error and exception dialogs."""

    def show_exception_dialog(
        self,
        parent: QWidget,
        text: str,
        informative_text: str,
        exception: Exception,
    ) -> None:
        """
        Show an error dialog describing an exception.

        Args:
            parent: a parent widget
            text: a main text
            informative_text: an informative text
            exception: an exception
        """
        self.show_error_dialog(
            parent=parent,
            text=text,
            informative_text=informative_text,
            detailed_text=error_to_string(
                error=exception,
            ),
        )

    def show_error_dialog(
        self,
        parent: QWidget,
        text: str,
        informative_text: str,
        detailed_text: str,
    ) -> None:
        """
        Show an error dialog.

        Args:
            parent: a parent widget
            text: a main text
            informative_text: an informative text
            detailed_text: a detailed text
        """
        dialog = QMessageBox(parent)
        dialog.setWindowTitle("ywh2bt - Error")
        dialog.setIcon(QMessageBox.Icon.Critical)
        dialog.setText(f"<b>{text}</b>")
        dialog.setInformativeText(informative_text)
        dialog.setDetailedText(detailed_text)
        dialog.show()



"""Models and functions used for YesWeHack data mapping."""
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Union,
    cast,
)

from yeswehack.api import CVSS as YesWeHackRawApiCvss
from yeswehack.api import Attachment as YesWeHackRawApiAttachment  # noqa: N811
from yeswehack.api import Author as YesWeHackRawApiAuthor
from yeswehack.api import BugType as YesWeHackRawApiBugType
from yeswehack.api import Log as YesWeHackRawApiLog
from yeswehack.api import Priority as YesWeHackRawApiPriority
from yeswehack.api import Report as YesWeHackRawApiReport

from ywh2bt.core.api.models.report import (
    AskForFixverificationStatusLog,
    Attachment,
    Author,
    BugType,
    CloseLog,
    CommentLog,
    Cvss,
    CvssUpdateLog,
    DetailsUpdateLog,
    FixVerifiedLog,
    Log,
    Priority,
    PriorityUpdateLog,
    Report,
    ReportProgram,
    RewardLog,
    StatusUpdateLog,
    TrackerMessageLog,
    TrackerUpdateLog,
    TrackingStatusLog,
    TransferLog,
)
from ywh2bt.core.html import (
    cleanup_attachments_and_urls_from_html,
    cleanup_ywh_redirects_from_html,
    cleanup_ywh_redirects_from_text,
)


@dataclass
class MappingContext:
    """A mapping context."""

    yeswehack_domain: str


def map_raw_report(
    context: MappingContext,
    raw_report: YesWeHackRawApiReport,
) -> Report:
    """
    Map a raw API report to a local report.

    Args:
        context: a mapping context
        raw_report: a raw report

    Returns:
        a local report
    """
    attachments = _map_raw_attachments(
        context=context,
        raw_attachments=raw_report.attachments,
    )
    description_html = cleanup_ywh_redirects_from_html(
        ywh_domain=context.yeswehack_domain,
        html=raw_report.description_html,
    )
    description_html, attachments = cleanup_attachments_and_urls_from_html(
        html=description_html,
        attachments=attachments,
    )
    bug_type = _map_raw_bug_type(
        context=context,
        raw_bug_type=raw_report.bug_type,
    )
    cvss = _map_raw_cvss(
        context=context,
        raw_cvss=raw_report.cvss,
    )
    priority = (
        _map_raw_priority(
            context=context,
            raw_priority=raw_report.priority,
        )
        if raw_report.priority
        else None
    )
    hunter = _map_raw_author(
        context=context,
        raw_author=raw_report.hunter,
    )
    status = _map_raw_status(
        context=context,
        raw_status=raw_report.status,
    )
    logs = map_raw_logs(
        context=context,
        raw_logs=raw_report.logs or [],
    )
    return Report(
        raw_report=raw_report,
        report_id=str(raw_report.id),
        title=raw_report.title,
        local_id=raw_report.local_id,
        bug_type=bug_type,
        scope=raw_report.scope,
        cvss=cvss,
        end_point=raw_report.end_point,
        vulnerable_part=raw_report.vulnerable_part,
        part_name=raw_report.part_name,
        payload_sample=raw_report.payload_sample,
        technical_environment=raw_report.technical_environment,
        description_html=description_html,
        attachments=attachments,
        hunter=hunter,
        status=status,
        tracking_status=raw_report.tracking_status,
        logs=logs,
        priority=priority,
        program=_map_raw_report_program(
            context=context,
            raw_program=raw_report.program or {},
        ),
        ask_for_fix_verification_status=raw_report.ask_for_fix_verification_status,
    )


def _map_raw_bug_type(
    context: MappingContext,
    raw_bug_type: YesWeHackRawApiBugType,
) -> BugType:
    return BugType(
        name=raw_bug_type.name,
        link=raw_bug_type.link,
        remediation_link=raw_bug_type.remediation_link,
    )


def _map_raw_priority(
    context: MappingContext,
    raw_priority: YesWeHackRawApiPriority,
) -> Priority:
    return Priority(
        name=raw_priority.name,
    )


def _map_raw_report_program(
    context: MappingContext,
    raw_program: Dict[Any, Any],
) -> ReportProgram:
    return ReportProgram(
        title=raw_program.get("title", ""),
        slug=raw_program.get("slug", ""),
    )


def _map_raw_cvss(
    context: MappingContext,
    raw_cvss: YesWeHackRawApiCvss,
) -> Cvss:
    return Cvss(
        criticity=raw_cvss.criticity,
        score=raw_cvss.score,
        vector=raw_cvss.vector,
    )


def _map_raw_attachments(
    context: MappingContext,
    raw_attachments: List[YesWeHackRawApiAttachment],
) -> List[Attachment]:
    return [
        map_raw_attachment(
            context=context,
            raw_attachment=raw_attachment,
        )
        for raw_attachment in raw_attachments
    ]


def map_raw_attachment(
    context: MappingContext,
    raw_attachment: YesWeHackRawApiAttachment,
) -> Attachment:
    """
    Map a raw API attachment to a local attachment.

    Args:
        context: a mapping context
        raw_attachment: a raw attachment

    Returns:
        a local attachment
    """
    return Attachment(
        attachment_id=raw_attachment.id,
        name=raw_attachment.name,
        original_name=raw_attachment.original_name,
        mime_type=raw_attachment.mime_type,
        size=raw_attachment.size,
        url=raw_attachment.url,
        data_loader=lambda: raw_attachment.data,
    )


def _map_raw_author(
    context: MappingContext,
    raw_author: Union[YesWeHackRawApiAuthor, Dict[str, Any]],
) -> Author:
    default_username = "Anonymous"
    if isinstance(raw_author, YesWeHackRawApiAuthor):
        username = raw_author.username or default_username
    else:
        username = raw_author.get("username", default_username)
    return Author(
        username=username,
    )


def _map_raw_status(
    context: MappingContext,
    raw_status: Dict[str, Any],
) -> str:
    return cast(str, raw_status.get("workflow_state", ""))


def map_raw_logs(
    context: MappingContext,
    raw_logs: List[YesWeHackRawApiLog],
) -> List[Log]:
    """
    Map a list of raw API logs to a list of local logs.

    Args:
        context: a mapping context
        raw_logs: a list of raw logs

    Returns:
        a list of local logs
    """
    return [
        map_raw_log(
            context=context,
            raw_log=raw_log,
        )
        for raw_log in raw_logs
    ]


def map_raw_log(
    context: MappingContext,
    raw_log: YesWeHackRawApiLog,
) -> Log:
    """
    Map a raw API log to a local log.

    Args:
        context: a mapping context
        raw_log: a raw log

    Returns:
        a local log
    """
    created_at = raw_log.created_at
    log_id = raw_log.id
    log_type = raw_log.type
    private = raw_log.private
    author = (
        _map_raw_author(
            context=context,
            raw_author=raw_log.author,
        )
        if raw_log.author
        else Author(
            username="Anonymous",
        )
    )
    attachments = _map_raw_attachments(
        context=context,
        raw_attachments=raw_log.attachments,
    )
    message_html = cleanup_ywh_redirects_from_html(
        ywh_domain=context.yeswehack_domain,
        html=raw_log.message_html or "",
    )
    message_html, attachments = cleanup_attachments_and_urls_from_html(
        html=message_html,
        attachments=attachments,
    )
    if raw_log.type == "close":
        return CloseLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_status=raw_log.old_status,
            new_status=raw_log.status,
        )
    if raw_log.type == "comment" or (raw_log.type == "assign" and message_html):
        return CommentLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
        )
    if raw_log.type == "cvss-update":
        return CvssUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_cvss=_map_raw_cvss(
                context=context,
                raw_cvss=raw_log.old_cvss,
            ),
            new_cvss=_map_raw_cvss(
                context=context,
                raw_cvss=raw_log.new_cvss,
            ),
        )
    if raw_log.type == "details-update":
        return DetailsUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_details=raw_log.old_details,
            new_details=raw_log.new_details,
        )
    if raw_log.type == "priority-update":
        return PriorityUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            new_priority=_map_raw_priority(
                context=context,
                raw_priority=raw_log.priority,
            )
            if raw_log.priority
            else None,
        )
    if raw_log.type == "reward":
        return RewardLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            reward_type=raw_log.reward_type,
        )
    if raw_log.type == "status-update":
        return StatusUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            old_status=raw_log.old_status,
            new_status=raw_log.status,
        )
    if raw_log.type == "tracking-status":
        return TrackingStatusLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            tracker_name=raw_log.tracker_name,
            tracker_url=cleanup_ywh_redirects_from_text(
                ywh_domain=context.yeswehack_domain,
                text=raw_log.tracker_url,
            )
            if raw_log.tracker_url
            else None,
            tracker_id=raw_log.tracker_id,
        )
    if raw_log.type == "tracker-update":
        return TrackerUpdateLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            tracker_name=raw_log.tracker_name,
            tracker_url=cleanup_ywh_redirects_from_text(
                ywh_domain=context.yeswehack_domain,
                text=raw_log.tracker_url,
            )
            if raw_log.tracker_url
            else None,
            tracker_id=raw_log.tracker_id,
            tracker_token=raw_log.tracker_token,
        )
    if raw_log.type == "tracker-message":
        return TrackerMessageLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            tracker_name=raw_log.tracker_name,
            tracker_url=cleanup_ywh_redirects_from_text(
                ywh_domain=context.yeswehack_domain,
                text=raw_log.tracker_url,
            )
            if raw_log.tracker_url
            else None,
            tracker_id=raw_log.tracker_id,
        )
    if raw_log.type == "fix-verified":
        return FixVerifiedLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            verified=raw_log.fix_verified,
        )
    if raw_log.type == "transfer":
        return TransferLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            program=raw_log.program,
            old_program=raw_log.old_program,
        )
    if raw_log.type == "ask-for-fix-verification-status":
        return AskForFixverificationStatusLog(
            created_at=created_at,
            log_id=log_id,
            log_type=log_type,
            private=private,
            author=author,
            message_html=message_html,
            attachments=attachments,
            new_ask_for_fix_verification_status=raw_log.new_ask_for_fix_verification_status,
            old_ask_for_fix_verification_status=raw_log.old_ask_for_fix_verification_status,
        )
    return Log(
        created_at=created_at,
        log_id=log_id,
        log_type=log_type,
        private=private,
        author=author,
        message_html=message_html,
        attachments=attachments,
    )



#!/usr/bin/env python
# coding: utf-8
# flake8: noqa
"""html2jira: Turn HTML into equivalent jira wiki format."""

# Based On https://github.com/Qazzian/html2jira

import html.entities as htmlentitydefs
import html.parser as HTMLParser
import re
import sys
import urllib.parse as urlparse
from textwrap import wrap


# Use Unicode characters instead of their ascii psuedo-replacements
UNICODE_SNOB = 0

# Escape all special characters.  Output is less readable, but avoids
# corner case formatting issues.
ESCAPE_SNOB = 0

# Put the links after each paragraph instead of at the end.
LINKS_EACH_PARAGRAPH = 0

# Wrap long lines at position. 0 for no wrapping. (Requires Python 2.3.)
BODY_WIDTH = 0

# Don't show internal links (href="#local-anchor") -- corresponding link
# targets won't be visible in the plain text file anyway.
SKIP_INTERNAL_LINKS = True

# Use inline, rather than reference, formatting for images and links
INLINE_LINKS = True

# Number of pixels Google indents nested lists
GOOGLE_LIST_INDENT = 36

IGNORE_ANCHORS = False
IGNORE_IMAGES = False
IGNORE_EMPHASIS = False

### Entity Nonsense ###
# For checking space-only lines on line 771
SPACE_RE = re.compile(r"\s\+")


def name2cp(k):
    if k == "apos":
        return ord("'")
    return htmlentitydefs.name2codepoint[k]


unifiable = {
    "rsquo": "'",
    "lsquo": "'",
    "rdquo": '"',
    "ldquo": '"',
    "copy": "(C)",
    "mdash": "--",
    "nbsp": " ",
    "rarr": "->",
    "larr": "<-",
    "middot": "*",
    "ndash": "-",
    "oelig": "oe",
    "aelig": "ae",
    "agrave": "a",
    "aacute": "a",
    "acirc": "a",
    "atilde": "a",
    "auml": "a",
    "aring": "a",
    "egrave": "e",
    "eacute": "e",
    "ecirc": "e",
    "euml": "e",
    "igrave": "i",
    "iacute": "i",
    "icirc": "i",
    "iuml": "i",
    "ograve": "o",
    "oacute": "o",
    "ocirc": "o",
    "otilde": "o",
    "ouml": "o",
    "ugrave": "u",
    "uacute": "u",
    "ucirc": "u",
    "uuml": "u",
    "lrm": "",
    "rlm": "",
}

unifiable_n = {}

for k in unifiable.keys():
    unifiable_n[name2cp(k)] = unifiable[k]


### End Entity Nonsense ###
def hn(tag):
    if tag[0] == "h" and len(tag) == 2:
        try:
            n = int(tag[1])
            if n in range(1, 10):
                return n
        except ValueError:
            return 0


def dumb_property_dict(style):
    """returns a hash of css attributes"""
    out = dict([(x.strip(), y.strip()) for x, y in [z.split(":", 1) for z in style.split(";") if ":" in z]])
    return out


def dumb_css_parser(data):
    """returns a hash of css selectors, each of which contains a hash of
    css attributes"""
    # remove @import sentences
    data += ";"
    importIndex = data.find("@import")
    while importIndex != -1:
        data = data[0:importIndex] + data[data.find(";", importIndex) + 1 :]
        importIndex = data.find("@import")

    # parse the css. reverted from dictionary compehension in order to
    # support older pythons
    elements = [x.split("{") for x in data.split("}") if "{" in x.strip()]
    try:
        elements = dict([(a.strip(), dumb_property_dict(b)) for a, b in elements])
    except ValueError:
        elements = {}  # not that important

    return elements


def element_style(attrs, style_def, parent_style):
    """returns a hash of the 'final' style attributes of the element"""
    style = parent_style.copy()
    if "class" in attrs:
        for css_class in attrs["class"].split():
            css_style = style_def["." + css_class]
            style.update(css_style)
    if "style" in attrs:
        immediate_style = dumb_property_dict(attrs["style"])
        style.update(immediate_style)
    return style


def google_list_style(style):
    """finds out whether this is an ordered or unordered list"""
    if "list-style-type" in style:
        list_style = style["list-style-type"]
        if list_style in ["disc", "circle", "square", "none"]:
            return "ul"
    return "ol"


def google_has_height(style):
    """check if the style of the element has the 'height' attribute
    explicitly defined"""
    if "height" in style:
        return True
    return False


def google_text_emphasis(style):
    """return a list of all emphasis modifiers of the element"""
    emphasis = []
    if "text-decoration" in style:
        emphasis.append(style["text-decoration"])
    if "font-style" in style:
        emphasis.append(style["font-style"])
    if "font-weight" in style:
        emphasis.append(style["font-weight"])
    return emphasis


def google_fixed_width_font(style):
    """check if the css of the current element defines a fixed width font"""
    font_family = ""
    if "font-family" in style:
        font_family = style["font-family"]
    if "Courier New" == font_family or "Consolas" == font_family:
        return True
    return False


def list_numbering_start(attrs):
    """extract numbering from list element attributes"""
    if "start" in attrs:
        try:
            return int(attrs["start"]) - 1
        except ValueError:
            pass

    return 0


class HTML2Jira(HTMLParser.HTMLParser):
    def __init__(self, out=None, baseurl="", bodywidth=BODY_WIDTH):
        HTMLParser.HTMLParser.__init__(self)

        # Config options
        self.unicode_snob = UNICODE_SNOB
        self.escape_snob = ESCAPE_SNOB
        self.links_each_paragraph = LINKS_EACH_PARAGRAPH
        self.body_width = bodywidth
        self.skip_internal_links = SKIP_INTERNAL_LINKS
        self.inline_links = INLINE_LINKS
        self.google_list_indent = GOOGLE_LIST_INDENT
        self.ignore_links = IGNORE_ANCHORS
        self.ignore_images = IGNORE_IMAGES
        self.ignore_emphasis = IGNORE_EMPHASIS
        self.google_doc = False
        self.ul_item_mark = "*"
        self.emphasis_mark = "_"
        self.strong_mark = "*"

        if out is None:
            self.out = self.outtextf
        else:
            self.out = out

        # empty list to store output characters before they are "joined"
        self.outtextlist = []

        self.outtext = str()

        self.quiet = 0
        self.p_p = 0  # number of newline character to print before next output
        self.outcount = 0
        self.start = 1
        self.space = 0
        self.a = []
        self.astack = []
        self.maybe_automatic_link = None
        self.absolute_url_matcher = re.compile(r"^[a-zA-Z+]+://")
        self.acount = 0
        self.list = []
        self.blockquote = 0
        self.pre = 0
        self.startpre = 0
        self.code = False
        self.br_toggle = ""
        self.lastWasNL = 0
        self.lastWasList = False
        self.style = 0
        self.style_def = {}
        self.tag_stack = []
        self.emphasis = 0
        self.drop_white_space = 0
        self.inheader = False
        self.abbr_title = None  # current abbreviation definition
        self.abbr_data = None  # last inner HTML (for abbr being defined)
        self.abbr_list = {}  # stack of abbreviations to write later
        self.baseurl = baseurl

        try:
            del unifiable_n[name2cp("nbsp")]
        except KeyError:
            pass
        unifiable["nbsp"] = "&nbsp_place_holder;"

    def feed(self, data):
        data = data.replace("</' + 'script>", "</ignore>")
        HTMLParser.HTMLParser.feed(self, data)

    def handle(self, data):
        self.feed(data)
        self.feed("")
        return self.optwrap(self.close())

    def outtextf(self, s):
        self.outtextlist.append(s)
        if s:
            self.lastWasNL = s[-1] == "\n"

    def close(self):
        HTMLParser.HTMLParser.close(self)

        self.pbr()
        self.o("", 0, "end")

        self.outtext = self.outtext.join(self.outtextlist)
        if self.unicode_snob:
            try:
                nbsp = chr(name2cp("nbsp"))
            except NameError:
                nbsp = chr(name2cp("nbsp"))
        else:
            try:
                nbsp = chr(32)
            except NameError:
                nbsp = chr(32)
        try:
            self.outtext = self.outtext.replace(unicode("&nbsp_place_holder;"), nbsp)
        except NameError:
            self.outtext = self.outtext.replace("&nbsp_place_holder;", nbsp)

        return self.outtext

    def handle_charref(self, c):
        self.o(self.charref(c), 1)

    def handle_entityref(self, c):
        self.o(self.entityref(c), 1)

    def handle_starttag(self, tag, attrs):
        self.handle_tag(tag, attrs, 1)

    def handle_endtag(self, tag):
        self.handle_tag(tag, None, 0)

    def previousIndex(self, attrs):
        """returns the index of certain set of attributes (of a link) in the
        self.a list

        If the set of attributes is not found, returns None
        """
        if "href" not in attrs:
            return None

        i = -1
        for a in self.a:
            i += 1
            match = 0

            if ("href" in a) and a["href"] == attrs["href"]:
                if ("title" in a) or ("title" in attrs):
                    if ("title" in a) and ("title" in attrs) and a["title"] == attrs["title"]:
                        match = True
                else:
                    match = True

            if match:
                return i

    def drop_last(self, nLetters):
        if not self.quiet:
            self.outtext = self.outtext[:-nLetters]

    def handle_emphasis(self, start, tag_style, parent_style):
        """handles various text emphases"""
        tag_emphasis = google_text_emphasis(tag_style)
        parent_emphasis = google_text_emphasis(parent_style)

        # handle Google's text emphasis
        strikethrough = "line-through" in tag_emphasis and self.hide_strikethrough
        bold = "bold" in tag_emphasis and not "bold" in parent_emphasis
        italic = "italic" in tag_emphasis and not "italic" in parent_emphasis
        fixed = google_fixed_width_font(tag_style) and not google_fixed_width_font(parent_style) and not self.pre

        if start:
            # crossed-out text must be handled before other attributes
            # in order not to output qualifiers unnecessarily
            if bold or italic or fixed:
                self.emphasis += 1
            if strikethrough:
                self.quiet += 1
            if italic:
                self.o(self.emphasis_mark)
                self.drop_white_space += 1
            if bold:
                self.o(self.strong_mark)
                self.drop_white_space += 1
            if fixed:
                self.o("`")
                self.drop_white_space += 1
                self.code = True
        else:
            if bold or italic or fixed:
                # there must not be whitespace before closing emphasis mark
                self.emphasis -= 1
                self.space = 0
                self.outtext = self.outtext.rstrip()
            if fixed:
                if self.drop_white_space:
                    # empty emphasis, drop it
                    self.drop_last(1)
                    self.drop_white_space -= 1
                else:
                    self.o("`")
                self.code = False
            if bold:
                if self.drop_white_space:
                    # empty emphasis, drop it
                    self.drop_last(2)
                    self.drop_white_space -= 1
                else:
                    self.o(self.strong_mark)
            if italic:
                if self.drop_white_space:
                    # empty emphasis, drop it
                    self.drop_last(1)
                    self.drop_white_space -= 1
                else:
                    self.o(self.emphasis_mark)
            # space is only allowed after *all* emphasis marks
            if (bold or italic) and not self.emphasis:
                self.o(" ")
            if strikethrough:
                self.quiet -= 1

    def handle_tag(self, tag, attrs, start):
        # attrs = fixattrs(attrs)
        # attrs is None for endtags
        if attrs is None:
            attrs = {}
        else:
            attrs = dict(attrs)

        if self.google_doc:
            # the attrs parameter is empty for a closing tag. in addition, we
            # need the attributes of the parent nodes in order to get a
            # complete style description for the current element. we assume
            # that google docs export well formed html.
            parent_style = {}
            if start:
                if self.tag_stack:
                    parent_style = self.tag_stack[-1][2]
                tag_style = element_style(attrs, self.style_def, parent_style)
                self.tag_stack.append((tag, attrs, tag_style))
            else:
                dummy, attrs, tag_style = self.tag_stack.pop()
                if self.tag_stack:
                    parent_style = self.tag_stack[-1][2]

        if hn(tag):
            self.p()
            if start:
                self.inheader = True
                self.o("h" + str(hn(tag)) + ". ")
            else:
                self.inheader = False
                return  # prevent redundant emphasis marks on headers

        if tag in ["p", "div"]:
            if self.google_doc:
                if start and google_has_height(tag_style):
                    self.p()
                else:
                    self.soft_br()
            else:
                self.p()

        if tag == "br" and start and not self.list:
            self.o("  \n")

        if tag == "hr" and start:
            self.p()
            self.o("----")
            self.p()

        if tag in ["head", "style", "script"]:
            if start:
                self.quiet += 1
            else:
                self.quiet -= 1

        if tag == "style":
            if start:
                self.style += 1
            else:
                self.style -= 1

        if tag in ["body"]:
            self.quiet = 0  # sites like 9rules.com never close <head>

        if tag == "blockquote":
            if start:
                self.p()
                self.o("> ", 0, 1)
                self.start = 1
                self.blockquote += 1
            else:
                self.blockquote -= 1
                self.p()

        if tag in ["em", "i", "u"] and not self.ignore_emphasis:
            self.o(self.emphasis_mark)
        if tag in ["strong", "b"] and not self.ignore_emphasis:
            self.o(self.strong_mark)
        if tag in ["del", "strike", "s"]:
            if start:
                self.o("<" + tag + ">")
            else:
                self.o("</" + tag + ">")

        if self.google_doc:
            if not self.inheader:
                # handle some font attributes, but leave headers clean
                self.handle_emphasis(start, tag_style, parent_style)

        if tag in ["code", "tt"] and not self.pre:
            self.o("`")  # TODO: `` `this` ``
        if tag == "abbr":
            if start:
                self.abbr_title = None
                self.abbr_data = ""
                if "title" in attrs:
                    self.abbr_title = attrs["title"]
            else:
                if self.abbr_title is not None:
                    self.abbr_list[self.abbr_data] = self.abbr_title
                    self.abbr_title = None
                self.abbr_data = ""

        if tag == "a" and not self.ignore_links:
            if start:
                if ("href" in attrs) and not (self.skip_internal_links and attrs["href"].startswith("#")):
                    self.astack.append(attrs)
                    self.maybe_automatic_link = attrs["href"]
                else:
                    self.astack.append(None)
            else:
                if self.astack:
                    a = self.astack.pop()
                    if self.maybe_automatic_link:
                        self.maybe_automatic_link = None
                    elif a:
                        if self.inline_links:
                            self.o("|" + escape_md(a["href"]) + "]")
                        else:
                            i = self.previousIndex(a)
                            if i is not None:
                                a = self.a[i]
                            else:
                                self.acount += 1
                                a["count"] = self.acount
                                a["outcount"] = self.outcount
                                self.a.append(a)
                            self.o("][" + str(a["count"]) + "]")

        if tag == "img" and start and not self.ignore_images:
            if "src" in attrs:
                attrs["href"] = attrs["src"]
                alt = attrs.get("alt", "")
                self.o("![" + escape_md(alt) + "]")

                if self.inline_links:
                    self.o("(" + escape_md(attrs["href"]) + ")")
                else:
                    i = self.previousIndex(attrs)
                    if i is not None:
                        attrs = self.a[i]
                    else:
                        self.acount += 1
                        attrs["count"] = self.acount
                        attrs["outcount"] = self.outcount
                        self.a.append(attrs)
                    self.o("[" + str(attrs["count"]) + "]")

        if tag == "dl" and start:
            self.p()
        if tag == "dt" and not start:
            self.pbr()
        if tag == "dd" and start:
            self.o("    ")
        if tag == "dd" and not start:
            self.pbr()

        if tag in ["ol", "ul"]:
            # Google Docs create sub lists as top level lists
            if (not self.list) and (not self.lastWasList):
                self.p()
            if start:
                if self.google_doc:
                    list_style = google_list_style(tag_style)
                else:
                    list_style = tag
                numbering_start = list_numbering_start(attrs)
                self.list.append({"name": list_style, "num": numbering_start})
            else:
                if self.list:
                    self.list.pop()
            self.lastWasList = True
        else:
            self.lastWasList = False

        if tag == "li":
            self.pbr()
            if start:
                if self.list:
                    li = self.list[-1]
                else:
                    li = {"name": "ul", "num": 0}
                if self.google_doc:
                    nest_count = self.google_nest_count(tag_style)
                else:
                    nest_count = len(self.list)
                # TODO: line up <ol><li>s > 9 correctly.
                self.o("  ")
                if li["name"] == "ul":
                    self.o((self.ul_item_mark * nest_count) + " ")
                elif li["name"] == "ol":
                    li["num"] += 1
                    self.o(str(li["num"]) + ". ")
                self.start = 1

        if tag in ["table", "tr"] and start:
            self.p()
        if tag == "td":
            self.pbr()

        if tag == "pre":
            if start:
                self.startpre = 1
                self.pre = 1
            else:
                self.pre = 0
            self.p()

    def pbr(self):
        if self.p_p == 0:
            self.p_p = 1

    def p(self):
        self.p_p = 2

    def soft_br(self):
        self.pbr()
        self.br_toggle = "  "

    def o(self, data, puredata=0, force=0):
        if self.abbr_data is not None:
            self.abbr_data += data

        if not self.quiet:
            if self.google_doc:
                # prevent white space immediately after 'begin emphasis'
                # marks ('**' and '_')
                lstripped_data = data.lstrip()
                if self.drop_white_space and not (self.pre or self.code):
                    data = lstripped_data
                if lstripped_data != "":
                    self.drop_white_space = 0

            if puredata and not self.pre:
                # This is a very dangerous call ... it could mess up
                # all handling of &nbsp; when not handled properly
                # (see entityref)
                data = re.sub(r"\s+", r" ", data)
                if data and data[0] == " ":
                    self.space = 1
                    data = data[1:]
            if not data and not force:
                return

            if self.startpre:
                # self.out(" :") #TODO: not output when already one there
                if not data.startswith("\n"):  # <pre>stuff...
                    data = "\n" + data

            bq = ">" * self.blockquote
            if not (force and data and data[0] == ">") and self.blockquote:
                bq += " "

            if self.pre:
                if not self.list:
                    bq += "    "
                # else: list content is already partially indented
                for i in range(len(self.list)):
                    bq += "    "
                data = data.replace("\n", "\n" + bq)

            if self.startpre:
                self.startpre = 0
                if self.list:
                    # use existing initial indentation
                    data = data.lstrip("\n")

            if self.start:
                self.space = 0
                self.p_p = 0
                self.start = 0

            if force == "end":
                # It's the end.
                self.p_p = 0
                self.out("\n")
                self.space = 0

            if self.p_p:
                self.out((self.br_toggle + "\n" + bq) * self.p_p)
                self.space = 0
                self.br_toggle = ""

            if self.space:
                if not self.lastWasNL:
                    self.out(" ")
                self.space = 0

            if self.a and ((self.p_p == 2 and self.links_each_paragraph) or force == "end"):
                if force == "end":
                    self.out("\n")

                newa = []
                for link in self.a:
                    if self.outcount > link["outcount"]:
                        self.out("   [" + str(link["count"]) + "]: " + urlparse.urljoin(self.baseurl, link["href"]))
                        if "title" in link:
                            self.out(" (" + link["title"] + ")")
                        self.out("\n")
                    else:
                        newa.append(link)

                # Don't need an extra line when nothing was done.
                if self.a != newa:
                    self.out("\n")

                self.a = newa

            if self.abbr_list and force == "end":
                for abbr, definition in self.abbr_list.items():
                    self.out("  *[" + abbr + "]: " + definition + "\n")

            self.p_p = 0
            self.out(data)
            self.outcount += 1

    def handle_data(self, data):
        if r"\/script>" in data:
            self.quiet -= 1

        if self.style:
            self.style_def.update(dumb_css_parser(data))

        if not self.maybe_automatic_link is None:
            href = self.maybe_automatic_link
            if href == data and self.absolute_url_matcher.match(href):
                self.o("<" + data + ">")
                return
            else:
                self.o("[")
                self.maybe_automatic_link = None

        if not self.code and not self.pre:
            data = escape_md_section(data, snob=self.escape_snob)
        self.o(data, 1)

    def unknown_decl(self, data):
        pass

    def charref(self, name):
        if name[0] in ["x", "X"]:
            c = int(name[1:], 16)
        else:
            c = int(name)

        if not self.unicode_snob and c in unifiable_n.keys():
            return unifiable_n[c]
        else:
            try:
                return chr(c)
            except NameError:  # Python3
                return chr(c)

    def entityref(self, c):
        if not self.unicode_snob and c in unifiable.keys():
            return unifiable[c]
        else:
            try:
                name2cp(c)
            except KeyError:
                return "&" + c + ";"
            else:
                if c == "nbsp":
                    return unifiable[c]
                else:
                    try:
                        return chr(name2cp(c))
                    except NameError:  # Python3
                        return chr(name2cp(c))

    def replaceEntities(self, s):
        s = s.group(1)
        if s[0] == "#":
            return self.charref(s[1:])
        else:
            return self.entityref(s)

    r_unescape = re.compile(r"&(#?[xX]?(?:[0-9a-fA-F]+|\w{1,8}));")

    def unescape(self, s):
        return self.r_unescape.sub(self.replaceEntities, s)

    def google_nest_count(self, style):
        """calculate the nesting count of google doc lists"""
        nest_count = 0
        if "margin-left" in style:
            nest_count = int(style["margin-left"][:-2]) // self.google_list_indent
        return nest_count

    def optwrap(self, text):
        """Wrap all paragraphs in the provided text."""
        if not self.body_width:
            return text

        assert wrap, "Requires Python 2.3."
        result = ""
        newlines = 0
        for para in text.split("\n"):
            if len(para) > 0:
                if not skipwrap(para):
                    result += "\n".join(wrap(para, self.body_width))
                    if para.endswith("  "):
                        result += "  \n"
                        newlines = 1
                    else:
                        result += "\n\n"
                        newlines = 2
                else:
                    # Warning for the tempted!!!
                    # Be aware that obvious replacement of this with
                    # line.isspace()
                    # DOES NOT work! Explanations are welcome.
                    if not SPACE_RE.match(para):
                        result += para + "\n"
                        newlines = 1
            else:
                if newlines < 2:
                    result += "\n"
                    newlines += 1
        return result


ordered_list_matcher = re.compile(r"\d+\.\s")
unordered_list_matcher = re.compile(r"[-\*\+]\s")
md_chars_matcher = re.compile(r"([\\\[\]\(\)])")
md_chars_matcher_all = re.compile(r"([`\*_{}\[\]\(\)#!])")
md_dot_matcher = re.compile(
    r"""
    ^             # start of line
    (\s*\d+)      # optional whitespace and a number
    (\.)          # dot
    (?=\s)        # lookahead assert whitespace
    """,
    re.MULTILINE | re.VERBOSE,
)
md_plus_matcher = re.compile(
    r"""
    ^
    (\s*)
    (\+)
    (?=\s)
    """,
    flags=re.MULTILINE | re.VERBOSE,
)
md_dash_matcher = re.compile(
    r"""
    ^
    (\s*)
    (-)
    (?=\s|\-)     # followed by whitespace (bullet list, or spaced out hr)
                  # or another dash (header or hr)
    """,
    flags=re.MULTILINE | re.VERBOSE,
)
slash_chars = r"\`*_{}[]()#+-.!"
md_backslash_matcher = re.compile(
    r"""
    (\\)          # match one slash
    (?=[%s])      # followed by a char that requires escaping
    """
    % re.escape(slash_chars),
    flags=re.VERBOSE,
)


def skipwrap(para):
    # If the text begins with four spaces or one tab, it's a code block;
    # don't wrap
    if para[0:4] == "    " or para[0] == "\t":
        return True
    # If the text begins with only two "--", possibly preceded by
    # whitespace, that's an emdash; so wrap.
    stripped = para.lstrip()
    if stripped[0:2] == "--" and len(stripped) > 2 and stripped[2] != "-":
        return False
    # I'm not sure what this is for; I thought it was to detect lists,
    # but there's a <br>-inside-<span> case in one of the tests that
    # also depends upon it.
    if stripped[0:1] == "-" or stripped[0:1] == "*":
        return True
    # If the text begins with a single -, *, or +, followed by a space,
    # or an integer, followed by a ., followed by a space (in either
    # case optionally preceeded by whitespace), it's a list; don't wrap.
    if ordered_list_matcher.match(stripped) or unordered_list_matcher.match(stripped):
        return True
    return False


def wrapwrite(text):
    text = text.encode("utf-8")
    try:  # Python3
        sys.stdout.buffer.write(text)
    except AttributeError:
        sys.stdout.write(text)


def html2jira(html, baseurl="", bodywidth=BODY_WIDTH) -> str:
    h = HTML2Jira(baseurl=baseurl, bodywidth=bodywidth)
    return h.handle(html)


def unescape(s, unicode_snob=False):
    h = HTML2Jira()
    h.unicode_snob = unicode_snob
    return h.unescape(s)


def escape_md(text):
    """Escapes markdown-sensitive characters within other markdown
    constructs."""
    return md_chars_matcher.sub(r"\\\1", text)


def escape_md_section(text, snob=False):
    """Escapes markdown-sensitive characters across whole document sections."""
    text = md_backslash_matcher.sub(r"\\\1", text)
    if snob:
        text = md_chars_matcher_all.sub(r"\\\1", text)
    text = md_dot_matcher.sub(r"\1\\\2", text)
    text = md_plus_matcher.sub(r"\1\\\2", text)
    text = md_dash_matcher.sub(r"\1\\\2", text)
    return text



"""Models and functions used for check box with hint."""
from typing import (
    Any,
    Dict,
    Optional,
)

from PySide6.QtCore import (
    Qt,
    Signal,
)
from PySide6.QtWidgets import (
    QCheckBox,
    QHBoxLayout,
    QLabel,
    QWidget,
)

from ywh2bt.gui.widgets.typing import as_signal_instance


class HintedCheckBoxWidget(QWidget):
    """A widget consisting of a checkbox with a hint."""

    stateChanged: Signal = Signal(Qt.CheckState)  # noqa: N815

    _check_box: QCheckBox
    _label: QLabel

    _state_hints: Dict[Qt.CheckState, str]

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._state_hints = {}
        self._init_ui()

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._check_box = self._create_check_box()

        self._label = self._create_label()

        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._check_box)
        layout.addWidget(self._label)
        layout.addStretch(1)

    def _create_check_box(
        self,
    ) -> QCheckBox:
        widget = QCheckBox(self)
        as_signal_instance(widget.stateChanged).connect(
            self._on_check_box_state_changed,
        )
        return widget

    def _on_check_box_state_changed(
        self,
        state: int,
    ) -> None:
        states = {
            0: Qt.CheckState.Unchecked,
            1: Qt.CheckState.PartiallyChecked,
            2: Qt.CheckState.Checked,
        }
        self._update_label_for_state(
            state=states[state],
        )
        as_signal_instance(self.stateChanged).emit(states[state])

    def set_state_hint(
        self,
        state: Qt.CheckState,
        hint: str,
    ) -> None:
        """
        Set a hint for a CheckState.

        Args:
            state: a state
            hint: a hint
        """
        self._state_hints[state] = hint

    def _update_label_for_state(
        self,
        state: Qt.CheckState,
    ) -> None:
        self._label.setText(self._state_hints.get(state, ""))

    def _create_label(
        self,
    ) -> QLabel:
        return QLabel(self)

    def setTristate(  # noqa: N802
        self,
        tristate: bool,
    ) -> None:
        """
        Set whether the checkbox is a tri-state checkbox.

        Args:
            tristate: a flag indicating if the checkbox is a tri-state checkbox
        """
        self._check_box.setTristate(tristate)

    def setCheckState(  # noqa: N802
        self,
        state: Qt.CheckState,
    ) -> None:
        """
        Set the checkbox's check state.

        Args:
            state: a state
        """
        self._check_box.setCheckState(state)



"""Models used for HTTP headers."""

from ywh2bt.core.configuration.attribute import ExportableDict


class Headers(ExportableDict[str, str, str, str]):
    """A class describing a list of HTTP headers."""



"""Models and functions used for JSON schema."""
import json
from functools import partial
from string import Template
from typing import (
    Any,
    Callable,
    Dict,
    List,
)

from ywh2bt.core.schema.error import SchemaError
from ywh2bt.core.schema.json import root_configuration_as_json_schema


Json = Dict[str, Any]


def root_configuration_as_text() -> str:
    """
    Get a text schema for RootConfiguration.

    Raises:
        SchemaError: if an error occurred during the creation of the schema

    Returns:
        the schema
    """
    json_schema_str = root_configuration_as_json_schema()
    try:
        schema: Json = json.loads(json_schema_str)
    except TypeError as load_error:
        raise SchemaError("JSON load error") from load_error
    return "\n".join(
        _schema_to_str(
            schema=schema,
        ),
    )


def _schema_to_str(
    schema: Json,
) -> List[str]:
    ref = schema.get("$ref")
    if ref:
        return [
            f'Reference: {ref.replace("#/definitions/", "")}',
        ]
    any_of = schema.get("anyOf")
    if any_of:
        return _extract_any_of(
            any_of=any_of,
        )
    output = _extract_attributes(
        schema=schema,
        attributes={
            "title": "Title",
            "description": "Description",
            "type": "Type",
            "const": "Required constant value",
            "default": "Default value",
        },
    )
    output.extend(
        _extract_properties(
            schema=schema,
        ),
    )
    output.extend(
        _extract_properties_patterns(
            schema=schema,
        ),
    )
    output.extend(
        _extract_definitions(
            schema=schema,
        ),
    )
    output.extend(
        _extract_items(
            schema=schema,
        ),
    )

    return output


def _extract_attributes(
    schema: Json,
    attributes: Dict[str, str],
) -> List[str]:
    output = []
    for attribute, label in attributes.items():
        value = schema.get(attribute)
        if value is not None:
            output.append(f"{label}: {value}")
    return output


def _extract_properties(
    schema: Json,
) -> List[str]:
    required = schema.get("required", [])
    properties = []
    for property_name, property_schema in schema.get("properties", {}).items():
        properties.append(f"- {property_name}")
        properties.extend(
            _prefix_lines(
                "  ",
                _schema_to_str(
                    schema=property_schema,
                ),
            ),
        )
        if property_name in required:
            properties.append("  Required: True")
    if not properties:
        return []
    return [
        "Properties:",
        *_prefix_lines(
            "  ",
            properties,
        ),
    ]


def _extract_properties_patterns(
    schema: Json,
) -> List[str]:
    return _extract_list(
        schema=schema,
        attribute_name="patternProperties",
        block_label="Properties patterns:",
        item_name_template=Template("- Pattern: ${item_name}"),
    )


def _extract_definitions(
    schema: Json,
) -> List[str]:
    return _extract_list(
        schema=schema,
        attribute_name="definitions",
        block_label="Definitions:",
        item_name_template=Template("${item_name}:"),
    )


def _extract_items(
    schema: Json,
) -> List[str]:
    items = schema.get("items")
    if not items:
        return []
    return [
        "Items:",
        *_prefix_lines(
            "  ",
            _schema_to_str(
                schema=items,
            ),
        ),
    ]


def _extract_list(
    schema: Json,
    attribute_name: str,
    block_label: str,
    item_name_template: Template,
) -> List[str]:
    items = []
    for item_name, item_schema in schema.get(attribute_name, {}).items():
        items.append(item_name_template.substitute(item_name=item_name))
        items.extend(
            _prefix_lines(
                "  ",
                _schema_to_str(
                    schema=item_schema,
                ),
            ),
        )
    if not items:
        return []
    return [
        block_label,
        *_prefix_lines(
            "  ",
            items,
        ),
    ]


def _extract_any_of(
    any_of: List[Json],
) -> List[str]:
    output = [
        "Any of:",
    ]
    for any_of_entry in any_of:
        output.extend(
            _prefix_lines(
                "  - ",
                list(
                    _schema_to_str(
                        schema=any_of_entry,
                    )
                ),
            ),
        )
    return output


def _prefix(
    prefix: str,
    string: str,
) -> str:
    return f"{prefix}{string}"


def _prefix_lines(
    prefix: str,
    lines: List[str],
) -> List[str]:
    return list(
        map(
            _prefixer(prefix),
            lines,
        )
    )


def _prefixer(
    prefix: str,
) -> Callable[[str], str]:
    return partial(
        _prefix,
        prefix,
    )



"""Models and functions used for data serialization/deserialization."""
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    TextIO,
)

from ywh2bt.core.exceptions import CoreException


class SerDe(ABC):
    """Base abstract serializer/deserializer."""

    @abstractmethod
    def serialize_to_stream(
        self,
        data: Any,
        stream: TextIO,
    ) -> None:
        """
        Write serialized data into a stream.

        Args:
            data: data
            stream: a stream
        """

    @abstractmethod
    def deserialize_from_stream(
        self,
        stream: TextIO,
    ) -> Any:
        """
        Read and deserialize data from a stream.

        Args:
            stream: a stream
        """


class SerDeError(CoreException):
    """A serialization error."""



"""Version."""
import os

import pkg_resources
import tomlkit


try:
    __VERSION__ = pkg_resources.get_distribution("ywh2bt").version
except pkg_resources.DistributionNotFound:
    # fallback if project is not installed
    pyproject_path = os.path.join(
        os.path.dirname(__file__),
        "..",
        "pyproject.toml",
    )
    with open(pyproject_path) as pyproject_file:
        pyproject = pyproject_file.read()
    __VERSION__ = "Unknown"
    tool = tomlkit.api.parse(pyproject).get("tool")  # type: ignore
    if tool:
        poetry = tool.get("poetry")
        if poetry:
            __VERSION__ = poetry.get("version", "Unknown")



"""Models and functions used for attributes container dict GUI."""
import re
from dataclasses import dataclass
from functools import partial
from string import Template
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)

from PySide6.QtCore import (
    QFile,
    QSize,
    Qt,
    Signal,
)
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLayout,
    QTabWidget,
    QToolButton,
    QWidget,
)

from ywh2bt.core.configuration.attribute import (
    AttributesContainer,
    AttributesContainerDict,
)
from ywh2bt.core.configuration.subtypable import SubtypableMetaclass
from ywh2bt.gui.widgets import constants
from ywh2bt.gui.widgets.attribute.attributes_container_dict_entry import AttributesContainerDictEntry
from ywh2bt.gui.widgets.typing import as_signal_instance


T_AC = AttributesContainer
T_ACD = AttributesContainerDict[T_AC]

ICON_PATH_TEMPLATES: Tuple[Template, ...] = (
    Template(":/resources/icons/types/${type_name}-2x.png"),
    Template(":/resources/icons/types/${type_name}.png"),
)


@dataclass
class _ButtonDescription:
    type_name: str
    icon_type_name: str
    entry_args: Dict[str, Any]


class AttributesContainerDictWidget(QWidget):
    """Attributes container dict GUI."""

    container_dict_changed: Signal = Signal(AttributesContainerDict)
    _container_dict_class: Type[T_ACD]
    _container_dict: Optional[T_ACD]

    _empty_widget: QWidget
    _tab_widget: QTabWidget

    def __init__(
        self,
        container_dict_class: Type[T_ACD],
        container_dict: Optional[T_ACD] = None,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            container_dict_class: a concrete class extending AttributesContainerDict
            container_dict: an attributes container dict
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._container_dict_class = container_dict_class
        self._container_dict = None
        self._init_ui()
        self.set_container_dict(
            container_dict=container_dict,
        )

    def set_container_dict(
        self,
        container_dict: Optional[T_ACD],
    ) -> None:
        """
        Set the container dict.

        Args:
            container_dict: a container dict
        """
        self._container_dict = container_dict
        self._remove_all_tabs()
        if container_dict is None or not isinstance(container_dict, dict):
            return
        for name, container in container_dict.items():
            entry = AttributesContainerDictEntry(
                key=name,
                value=container,
            )
            self._add_tab(
                entry=entry,
            )

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._tab_widget = self._create_tab_widget()
        self._empty_widget = self._create_empty_widget()

        self._auto_show_hide_tabs()

        layout.addWidget(self._empty_widget)
        layout.addWidget(self._tab_widget)
        layout.setContentsMargins(0, 0, 0, 0)

    def _create_empty_widget(
        self,
    ) -> QWidget:
        return self._create_buttons_widget(
            text_template=Template("Add"),
            tool_tip_template=Template("Add ${type_name}"),
            buttons_size=constants.BIG_BUTTON_SIZE,
            icons_size=constants.BIG_BUTTON_ICON_SIZE,
        )

    def _create_tab_widget(
        self,
    ) -> QTabWidget:
        corner_widget = self._create_buttons_widget(
            text_template=Template(""),
            tool_tip_template=Template("Add ${type_name}"),
            buttons_size=constants.TAB_WIDGET_CORNER_BUTTON_SIZE,
            icons_size=constants.SMALL_BUTTON_ICON_SIZE,
        )

        widget = QTabWidget(self)
        widget.setCornerWidget(corner_widget)
        widget.setTabsClosable(True)
        as_signal_instance(widget.tabCloseRequested).connect(
            self._on_tab_close_requested,
        )
        return widget

    def _create_buttons_widget(
        self,
        text_template: Template,
        tool_tip_template: Template,
        buttons_size: QSize,
        icons_size: QSize,
    ) -> QWidget:
        layout = self._create_buttons_layout(
            text_template=text_template,
            tool_tip_template=tool_tip_template,
            buttons_size=buttons_size,
            icons_size=icons_size,
        )
        margin_left, margin_top, margin_right, margin_bottom = cast(
            Tuple[int, int, int, int], layout.getContentsMargins()
        )
        layout.setContentsMargins(0, margin_top, margin_right, margin_bottom)
        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def _create_add_button(
        self,
        icon: Optional[QIcon],
        text: Optional[str],
        tool_tip: Optional[str],
        entry_args: Dict[str, Any],
        size: QSize,
        icon_size: QSize,
    ) -> QToolButton:
        widget = QToolButton(self)
        widget.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        if text:
            widget.setText(text)
        if icon:
            widget.setIcon(icon)
        if tool_tip:
            widget.setToolTip(tool_tip)
        widget.setFixedSize(size)
        widget.setIconSize(icon_size)
        as_signal_instance(widget.clicked).connect(
            partial(
                self._on_add_button_clicked,
                entry_args,
            ),
        )
        return widget

    def _get_button_descriptions(
        self,
    ) -> List[_ButtonDescription]:
        button_descriptions: List[_ButtonDescription] = []
        item_values_type = self._container_dict_class().values_type  # type: ignore
        if isinstance(item_values_type, SubtypableMetaclass):
            values_type = cast(SubtypableMetaclass, item_values_type)
            types = values_type.get_registered_subtypes()
            for type_name, type_class in types.items():
                button_descriptions.append(
                    _ButtonDescription(
                        type_name=type_name,
                        icon_type_name=f"{values_type.__name__}/{type_class.__name__}",
                        entry_args={
                            "type": type_name,
                        },
                    ),
                )
        else:
            button_descriptions.append(
                _ButtonDescription(
                    type_name=item_values_type.__name__,
                    icon_type_name=item_values_type.__name__,
                    entry_args={},
                ),
            )
        return button_descriptions

    def _create_buttons_layout(
        self,
        text_template: Template,
        tool_tip_template: Template,
        buttons_size: QSize,
        icons_size: QSize,
    ) -> QLayout:
        layout = QHBoxLayout()
        descriptions = self._get_button_descriptions()
        for description in descriptions:
            button = self._create_add_button(
                icon=self._get_type_icon(
                    type_name=description.icon_type_name,
                ),
                text=text_template.substitute(
                    type_name=description.type_name,
                ),
                tool_tip=tool_tip_template.substitute(
                    type_name=description.type_name,
                ),
                entry_args=description.entry_args,
                size=buttons_size,
                icon_size=icons_size,
            )
            layout.addWidget(button, 1)
        layout.addStretch(1)
        return layout

    def _get_type_icon(
        self,
        type_name: str,
    ) -> Optional[QIcon]:
        paths = [
            template.substitute(
                type_name=type_name,
            )
            for template in ICON_PATH_TEMPLATES
        ]
        return _try_get_icon(
            try_paths=paths,
        )

    def _on_add_button_clicked(
        self,
        entry_args: Dict[str, Any],
    ) -> None:
        entry = self._create_container_dict_entry(
            entry_args=entry_args,
        )
        new_tab_index = self._add_tab(
            entry=entry,
        )
        self._tab_widget.setCurrentIndex(new_tab_index)

    def _remove_all_tabs(
        self,
    ) -> None:
        self._tab_widget.blockSignals(True)
        while self._tab_widget.count():
            self._tab_widget.removeTab(0)
        self._tab_widget.blockSignals(False)
        self._auto_show_hide_tabs()

    def _add_tab(
        self,
        entry: AttributesContainerDictEntry,
    ) -> int:
        from ywh2bt.gui.widgets.attribute.attributes_container_dict_entry_widget import (  # noqa: N400
            AttributesContainerDictEntryWidget,
        )

        widget = AttributesContainerDictEntryWidget(
            entry=entry,
        )
        as_signal_instance(widget.key_changed).connect(
            self._on_container_dict_entry_key_changed,
        )
        as_signal_instance(widget.value_changed).connect(
            self._on_container_dict_entry_value_changed,
        )
        icon = self._get_entry_tab_icon(entry)
        if icon:
            tab_index = self._tab_widget.addTab(
                widget,
                icon,
                entry.key,
            )
        else:
            tab_index = self._tab_widget.addTab(
                widget,
                entry.key,
            )
        self._auto_show_hide_tabs()
        return tab_index

    def _get_entry_tab_icon(
        self,
        entry: AttributesContainerDictEntry,
    ) -> Optional[QIcon]:
        item_values_type = self._container_dict_class().values_type  # type: ignore
        if isinstance(item_values_type, SubtypableMetaclass):
            values_type = cast(SubtypableMetaclass, item_values_type)
            types = values_type.get_registered_subtypes()
            for type_name, type_class in types.items():
                if isinstance(entry.value, type_class):
                    return self._get_type_icon(f"{values_type.__name__}/{type_class.__name__}")
        else:
            return self._get_type_icon(item_values_type.__name__)
        return None

    def _on_tab_close_requested(
        self,
        tab_index: int,
    ) -> None:
        if self._container_dict:
            keys = list(self._container_dict.keys())
            self._delete_container_dict_entry(
                key=keys[tab_index],
            )
            self._tab_widget.removeTab(tab_index)
        self._auto_show_hide_tabs()

    def _auto_show_hide_tabs(
        self,
    ) -> None:
        if self._tab_widget.count() == 0:
            self._tab_widget.hide()
            self._empty_widget.show()
        else:
            self._tab_widget.show()
            self._empty_widget.hide()

    def _on_container_dict_entry_key_changed(
        self,
        entry: AttributesContainerDictEntry,
    ) -> None:
        if self._container_dict is None:
            return
        key_index = None
        for index, item in enumerate(self._container_dict.items()):
            key, value = item
            if value == entry.value:
                key_index = (key, index)
                break
        if not key_index:
            raise KeyError(f"Could not find old key for {entry}")
        self._container_dict.swap_key(
            old=key_index[0],
            new=entry.key,
        )
        self._tab_widget.setTabText(
            key_index[1],
            entry.key,
        )
        self._emit_container_dict_changed_signal()

    def _on_container_dict_entry_value_changed(
        self,
        entry: AttributesContainerDictEntry,
    ) -> None:
        if self._container_dict is None:
            return
        self._container_dict[entry.key] = entry.value
        self._emit_container_dict_changed_signal()

    def _create_container_dict_entry(
        self,
        entry_args: Dict[str, Any],
    ) -> AttributesContainerDictEntry:
        if not self._container_dict:
            self._container_dict = self._container_dict_class()  # type: ignore
        item_values_type = self._container_dict.values_type
        value = item_values_type(**entry_args)
        key = _get_next_available_key(
            container_dict=self._container_dict,
            value=value,
        )
        self._container_dict[key] = value
        self._emit_container_dict_changed_signal()
        return AttributesContainerDictEntry(
            key=key,
            value=value,
        )

    def _delete_container_dict_entry(
        self,
        key: str,
    ) -> None:
        if self._container_dict is None:
            return
        self._container_dict.pop(key)
        self._emit_container_dict_changed_signal()

    def _emit_container_dict_changed_signal(
        self,
    ) -> None:
        as_signal_instance(self.container_dict_changed).emit(
            self._container_dict,
        )


def _get_icon(
    path: str,
) -> Optional[QIcon]:
    if QFile.exists(path):
        return QIcon(path)
    return None


def _value_class_to_snake_case(
    value: Any,
) -> str:
    pattern = re.compile("(?<!^)(?=[A-Z])")
    return pattern.sub("_", value.__class__.__name__).lower()


def _try_get_icon(
    try_paths: List[str],
) -> Optional[QIcon]:
    for try_path in try_paths:
        icon = _get_icon(
            path=try_path,
        )
        if icon:
            return icon
    return None


def _get_next_available_key(
    container_dict: Optional[T_ACD],
    value: AttributesContainer,
) -> str:
    i = 1
    while True:
        snake = _value_class_to_snake_case(
            value=value,
        )
        key = f"{snake}_{i}"
        if not container_dict or key not in container_dict:
            return key
        i += 1



"""Models and functions used for base api clients."""
from abc import (
    ABC,
    abstractmethod,
)


class TestableApiClient(ABC):
    """An API client that can be tested."""

    @abstractmethod
    def test(
        self,
    ) -> None:
        """Test the client."""



"""Models and functions used for the main GUI."""
import argparse
import platform
import sys

import urllib3


try:
    from PySide6.QtGui import QIcon
    from PySide6.QtWidgets import QApplication
except ModuleNotFoundError:
    print(
        "PySide6 package not found. Did you forget to install 'gui' extra?\n -> pip install 'ywh2bt[gui]'",
        file=sys.stderr,
    )
    sys.exit(1)

# noinspection PyUnresolvedReferences
import ywh2bt.gui.resources  # noqa: F401
from ywh2bt.core.core import AVAILABLE_FORMATS
from ywh2bt.gui.settings import settings
from ywh2bt.gui.widgets.main_window import MainWindow
from ywh2bt.version import __VERSION__


if platform.system() == "Windows":
    import ctypes

urllib3.disable_warnings(
    category=urllib3.exceptions.InsecureRequestWarning,
)

DEFAULT_FORMATTER_CLASS = argparse.ArgumentDefaultsHelpFormatter
CONFIGURATION_FORMATS = list(AVAILABLE_FORMATS.keys())


def run(*args: str) -> None:
    """
    Start the main GUI.

    Args:
        args: command line arguments
    """
    if not args:
        args = tuple(sys.argv[1:])
    parsed = _parse_args(*args)

    argv = [
        "ywh2bt-gui",
        "-qwindowtitle",
        "ywh2bt",
    ]
    app = QApplication(argv)

    main_icon = QIcon(":/resources/icons/ywh2bt.png")
    app.setWindowIcon(main_icon)

    system = platform.system()
    if system == "Windows":
        # set explicit AppUserModelID ; allows custom icon in Windows taskbar
        # see https://stackoverflow.com/a/1552105/248343
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(  # type: ignore
            f"yeswehack.ywh2bt.gui.{__VERSION__}".encode("utf-8"),
        )

    main_window = MainWindow(
        settings=settings,
    )
    main_window.setObjectName("main_window")
    main_window.setWindowIcon(main_icon)
    main_window.setWindowTitle("ywh2bt")
    main_window.show()

    if parsed.config_file:
        main_window.open_file(
            file_path=parsed.config_file,
            file_format=parsed.config_format,
        )

    sys.exit(app.exec())


def _parse_args(
    *args: str,
) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Open ywh2bt GUI",
        formatter_class=DEFAULT_FORMATTER_CLASS,
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__VERSION__} (python {'.'.join(map(str, sys.version_info[:3]))})",  # noqa: C812
    )
    parser.add_argument(
        "--config-file",
        "-c",
        dest="config_file",
        help="path to configuration file to be opened at GUI start-up",
        required=False,
        type=str,
    )
    parser.add_argument(
        "--config-format",
        "-f",
        dest="config_format",
        help="format of configuration file",
        type=str,
        required=False,
        default="yaml",
        choices=CONFIGURATION_FORMATS,
    )

    return parser.parse_args(
        args,
    )


if __name__ == "__main__":
    run(*sys.argv[1:])



"""Models and functions used for attributes container list GUI."""
from typing import (
    Any,
    Optional,
    Type,
)

from PySide6.QtCore import Signal
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLayout,
    QPushButton,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from ywh2bt.core.configuration.attribute import (
    AttributesContainer,
    AttributesContainerList,
)
from ywh2bt.gui.widgets import constants
from ywh2bt.gui.widgets.tab_widget_helper import object_to_tab_title
from ywh2bt.gui.widgets.typing import as_signal_instance


T_AC = AttributesContainer
T_ACL = AttributesContainerList[T_AC]


class AttributesContainerListWidget(QWidget):
    """Attributes container list GUI."""

    container_list_changed: Signal = Signal(AttributesContainerList)
    _container_list_class: Type[T_ACL]
    _container_list: Optional[T_ACL]

    _empty_widget: QWidget
    _tab_widget: QTabWidget

    def __init__(
        self,
        container_list_class: Type[T_ACL],
        container_list: Optional[T_ACL] = None,
        parent: Optional[QWidget] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Initialize self.

        Args:
            container_list_class: a concrete class extending AttributesContainerList
            container_list: an attributes container list
            parent: a parent widget
            args: extra arguments
            kwargs: extra keyword arguments
        """
        super().__init__(
            parent,
            *args,
            **kwargs,
        )
        self._container_list_class = container_list_class
        self._container_list = None
        self._init_ui()
        self.set_container_list(
            container_list=container_list,
        )

    def set_container_list(
        self,
        container_list: Optional[T_ACL],
    ) -> None:
        """
        Set the container list.

        Args:
            container_list: a container list
        """
        self._container_list = container_list
        if container_list is None:
            return
        for container in container_list:
            self._add_tab(
                container=container,
            )

    def _init_ui(self) -> None:
        layout = QHBoxLayout(self)

        self._tab_widget = self._create_tab_widget()
        self._empty_widget = self._create_empty_widget()

        layout.addWidget(self._empty_widget)
        layout.addWidget(self._tab_widget)
        layout.setContentsMargins(0, 0, 0, 0)

        self._auto_show_hide_tabs()

    def _create_empty_widget(
        self,
    ) -> QWidget:
        return self._create_buttons_widget()

    def _create_tab_widget(
        self,
    ) -> QTabWidget:
        widget = QTabWidget(self)
        widget.setTabsClosable(True)
        widget.setCornerWidget(self._create_buttons_widget())
        as_signal_instance(widget.tabCloseRequested).connect(
            self._on_tab_close_requested,
        )
        return widget

    def _create_buttons_widget(
        self,
    ) -> QWidget:
        button = QPushButton(
            "+",
            self,
        )
        button.setFixedSize(constants.SMALL_BUTTON_SIZE)
        as_signal_instance(button.clicked).connect(
            self._on_add_button_clicked,
        )
        layout = QHBoxLayout()
        layout.addWidget(button, 1)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def _create_buttons_layout(
        self,
        *buttons: QPushButton,
    ) -> QLayout:
        layout = QVBoxLayout()
        for button in buttons:
            layout.addWidget(button, 1)
        layout.addStretch(1)
        return layout

    def _on_add_button_clicked(
        self,
        checked: bool,
    ) -> None:
        new_entry = self._create_container_list_entry()
        new_tab_index = self._add_tab(
            container=new_entry,
        )
        self._tab_widget.setCurrentIndex(new_tab_index)

    def _auto_show_hide_tabs(
        self,
    ) -> None:
        if self._tab_widget.count() == 0:
            self._tab_widget.hide()
            self._empty_widget.show()
        else:
            self._tab_widget.show()
            self._empty_widget.hide()

    def _add_tab(
        self,
        container: AttributesContainer,
    ) -> int:
        from ywh2bt.gui.widgets.attribute.attributes_container_widget import AttributesContainerWidget

        widget = AttributesContainerWidget(
            parent=self,
            container_class=container.__class__,
            container=container,
        )
        as_signal_instance(widget.container_changed).connect(
            self._on_attributes_container_widget_changed,
        )
        title = self._attributes_container_to_tab_title(
            container=container,
        )
        tab_index = self._tab_widget.addTab(
            widget,
            title,
        )
        self._auto_show_hide_tabs()
        return tab_index

    def _on_tab_close_requested(
        self,
        tab_index: int,
    ) -> None:
        self._delete_container_list_entry(
            index=tab_index,
        )
        self._tab_widget.removeTab(tab_index)

        self._auto_show_hide_tabs()

    def _attributes_container_to_tab_title(
        self,
        container: AttributesContainer,
    ) -> str:
        if self._container_list is None:
            return "Tab"
        title = object_to_tab_title(
            obj=container,
        )
        if not title:
            index = self._container_list.index(container)
            class_name = self._container_list.values_type.__name__
            position = index + 1
            title = f"{class_name} #{position}"
        return title

    def _on_attributes_container_widget_changed(
        self,
        container: AttributesContainer,
    ) -> None:
        if self._container_list is None:
            return
        index = self._container_list.index(container)
        self._tab_widget.setTabText(
            index,
            self._attributes_container_to_tab_title(
                container=container,
            ),
        )
        self._emit_container_list_changed_signal()

    def _create_container_list_entry(
        self,
    ) -> AttributesContainer:
        if self._container_list is None:
            self._container_list = self._container_list_class()  # type: ignore
        item_values_type = self._container_list.values_type
        entry = item_values_type()
        self._container_list.append(entry)
        self._emit_container_list_changed_signal()
        return entry

    def _delete_container_list_entry(
        self,
        index: int,
    ) -> None:
        if self._container_list is None:
            return
        self._container_list.pop(index)
        self._emit_container_list_changed_signal()

    def _emit_container_list_changed_signal(
        self,
    ) -> None:
        as_signal_instance(self.container_list_changed).emit(
            self._container_list,
        )
